from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw
from narwhals.exceptions import MultiOutputExpressionError
from tests.utils import (
    DUCKDB_VERSION,
    Constructor,
    ConstructorEager,
    assert_equal_data,
    uses_pyarrow_backend,
)

if TYPE_CHECKING:
    from narwhals.typing import _1DArray

data = {
    "a": [1, 2, 3],
    "b": ["a", "b", "c"],
    "c": [4.1, 5.0, 6.0],
    "d": [True, False, True],
    "e": [7.0, 2.0, 1.1],
}


def test_when(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(nw.when(nw.col("a") == 1).then(value=3).alias("a_when"))
    expected = {"a_when": [3, None, None]}
    assert_equal_data(result, expected)
    result = df.select(nw.when(nw.col("a") == 1).then(value=3))
    expected = {"literal": [3, None, None]}
    assert_equal_data(result, expected)


def test_when_otherwise(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(nw.when(nw.col("a") == 1).then(3).otherwise(6).alias("a_when"))
    expected = {"a_when": [3, 6, 6]}
    assert_equal_data(result, expected)


def test_multiple_conditions(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(
        nw.when(nw.col("a") < 3, nw.col("c") < 5.0).then(3).alias("a_when")
    )
    expected = {"a_when": [3, None, None]}
    assert_equal_data(result, expected)


def test_no_arg_when_fail(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    with pytest.raises((TypeError, ValueError)):
        df.select(nw.when().then(value=3).alias("a_when"))


def test_value_numpy_array(constructor_eager: ConstructorEager) -> None:
    pytest.importorskip("numpy")
    import numpy as np

    df = nw.from_native(constructor_eager(data))

    result = df.select(nw.when(nw.col("a") == 1).then(np.arange(3, 6)).alias("a_when"))
    expected = {"a_when": [3, None, None]}
    assert_equal_data(result, expected)


def test_value_series(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager(data))
    s_data = {"s": [3, 4, 5]}
    s = nw.from_native(constructor_eager(s_data))["s"]
    assert isinstance(s, nw.Series)
    result = df.select(nw.when(nw.col("a") == 1).then(s).alias("a_when"))
    expected = {"a_when": [3, None, None]}
    assert_equal_data(result, expected)


def test_value_expression(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(nw.when(nw.col("a") == 1).then(nw.col("a") + 9).alias("a_when"))
    expected = {"a_when": [10, None, None]}
    assert_equal_data(result, expected)


def test_otherwise_numpy_array(constructor_eager: ConstructorEager) -> None:
    pytest.importorskip("numpy")
    import numpy as np

    df = nw.from_native(constructor_eager(data))

    arr: _1DArray = np.zeros([3], np.dtype(np.int64))
    arr[:3] = 0, 9, 10
    result = df.select(nw.when(nw.col("a") == 1).then(-1).otherwise(arr).alias("a_when"))
    expected = {"a_when": [-1, 9, 10]}
    assert_equal_data(result, expected)


def test_otherwise_series(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager(data))
    s_data = {"s": [0, 9, 10]}
    s = nw.from_native(constructor_eager(s_data))["s"]
    assert isinstance(s, nw.Series)
    result = df.select(nw.when(nw.col("a") == 1).then(-1).otherwise(s).alias("a_when"))
    expected = {"a_when": [-1, 9, 10]}
    assert_equal_data(result, expected)


def test_otherwise_expression(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(
        nw.when(nw.col("a") == 1).then(-1).otherwise(nw.col("a") + 7).alias("a_when")
    )
    expected = {"a_when": [-1, 9, 10]}
    assert_equal_data(result, expected)


def test_when_then_otherwise_into_expr(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(nw.when(nw.col("a") > 1).then("c").otherwise("e"))
    expected = {"c": [7, 5, 6]}
    assert_equal_data(result, expected)


def test_when_then_broadcasting(constructor: Constructor) -> None:
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    df = nw.from_native(constructor(data))
    result = df.select(nw.when(nw.col("a").sum() > 1).then("c"))
    expected = {"c": [4.1, 5, 6]}
    assert_equal_data(result, expected)
    result = df.select(nw.when(nw.col("a").sum() > 1).then(1).otherwise("c"))
    expected = {"literal": [1, 1, 1]}
    assert_equal_data(result, expected)


def test_when_then_otherwise_lit_str(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(nw.when(nw.col("a") > 1).then(nw.col("b")).otherwise(nw.lit("z")))
    expected = {"b": ["z", "b", "c"]}
    assert_equal_data(result, expected)


def test_when_then_otherwise_both_lit(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(
        x1=nw.when(nw.col("a") > 1).then(nw.lit(42)).otherwise(nw.lit(-1)),
        x2=nw.when(nw.col("a") > 2).then(nw.lit(42)).otherwise(nw.lit(-1)),
    )
    expected = {"x1": [-1, 42, 42], "x2": [-1, -1, 42]}
    assert_equal_data(result, expected)


def test_when_then_otherwise_multi_output(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3], "b": [4, 5, 6]}))
    with pytest.raises(MultiOutputExpressionError):
        df.select(x1=nw.when(nw.all() > 1).then(nw.col("a", "b")))
    with pytest.raises(MultiOutputExpressionError):
        df.select(x1=nw.when(nw.all() > 1).then(nw.lit(1)).otherwise(nw.all()))


@pytest.mark.parametrize(
    ("condition", "then", "otherwise", "expected"),
    [
        (nw.col("a").sum() == 6, 100, None, [100]),
        (nw.col("a").sum() == 6, 100, 200, [100]),
        (nw.col("a").sum() == 6, nw.col("a").sum(), 200, [6]),
        (nw.col("a").sum() == 6, 100, nw.col("b").sum(), [100]),
        (nw.col("a").sum() == 6, nw.col("a").sum(), nw.col("b").sum(), [6]),
        (nw.col("a").sum() == 5, 100, None, [None]),
        (nw.col("a").sum() == 5, 100, 200, [200]),
        (nw.col("a").sum() == 5, nw.col("a").sum(), 200, [200]),
        (nw.col("a").sum() == 5, 100, nw.col("b").sum(), [15]),
        (nw.col("a").sum() == 5, nw.col("a").sum(), nw.col("b").sum(), [15]),
    ],
)
def test_when_then_otherwise_aggregate_select(
    condition: nw.Expr,
    then: nw.Expr | int,
    otherwise: nw.Expr | int | None,
    expected: list[int],
    constructor: Constructor,
    request: pytest.FixtureRequest,
) -> None:
    if "cudf" in str(constructor) and otherwise is None:
        reason = "cudf does not support mixed types"
        request.applymarker(pytest.mark.xfail(reason=reason))
    df = nw.from_native(constructor({"a": [1, 2, 3], "b": [4, 5, 6]}))
    result = df.select(a_when=nw.when(condition).then(then).otherwise(otherwise))
    assert_equal_data(result, {"a_when": expected})


@pytest.mark.parametrize(
    ("condition", "then", "otherwise", "expected"),
    [
        (nw.col("a").sum() == 6, 100, None, [100, 100, 100]),
        (nw.col("a").sum() == 6, 100, 200, [100, 100, 100]),
        (nw.col("a").sum() == 6, nw.col("a").sum(), 200, [6, 6, 6]),
        (nw.col("a").sum() == 6, 100, nw.col("b").sum(), [100, 100, 100]),
        (nw.col("a").sum() == 6, nw.col("a").sum(), nw.col("b").sum(), [6, 6, 6]),
        (nw.col("a").sum() == 5, 100, None, [None, None, None]),
        (nw.col("a").sum() == 5, 100, 200, [200, 200, 200]),
        (nw.col("a").sum() == 5, nw.col("a").sum(), 200, [200, 200, 200]),
        (nw.col("a").sum() == 5, 100, nw.col("b").sum(), [15, 15, 15]),
        (nw.col("a").sum() == 5, nw.col("a").sum(), nw.col("b").sum(), [15, 15, 15]),
    ],
)
def test_when_then_otherwise_aggregate_with_columns(
    condition: nw.Expr,
    then: nw.Expr | int,
    otherwise: nw.Expr | int | None,
    expected: list[int],
    constructor: Constructor,
    request: pytest.FixtureRequest,
) -> None:
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    if "cudf" in str(constructor) and otherwise is None:
        reason = "cudf does not support mixed types"
        request.applymarker(pytest.mark.xfail(reason=reason))

    df = nw.from_native(constructor({"a": [1, 2, 3], "b": [4, 5, 6]}))
    expr = nw.when(condition).then(then).otherwise(otherwise)
    result = df.with_columns(a_when=expr)
    assert_equal_data(result.select(nw.col("a_when")), {"a_when": expected})


def test_when_then_empty(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [-1]})).filter(nw.col("a") > 0)
    result = df.with_columns(nw.when(nw.col("a") == 1).then(nw.lit(1)).alias("new_col"))
    expected: dict[str, Any] = {"a": [], "new_col": []}
    assert_equal_data(result, expected)


def test_when_chain_basic_two_conditions(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3]}))

    result = df.select(
        nw.when(nw.col("a") == 1)
        .then(10)
        .when(nw.col("a") == 2)
        .then(20)
        .otherwise(30)
        .alias("result")
    )

    expected = {"result": [10, 20, 30]}
    assert_equal_data(result, expected)


def test_when_chain_three_conditions(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3, 4]}))

    result = df.select(
        nw.when(nw.col("a") == 1)
        .then(100)
        .when(nw.col("a") == 2)
        .then(200)
        .when(nw.col("a") == 3)
        .then(300)
        .otherwise(400)
        .alias("result")
    )

    expected = {"result": [100, 200, 300, 400]}
    assert_equal_data(result, expected)


def test_when_chain_multiple_conditions(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3, 4, 5]}))

    result = df.select(
        nw.when(nw.col("a") == 1)
        .then(10)
        .when(nw.col("a") == 2)
        .then(20)
        .when(nw.col("a") == 3)
        .then(30)
        .when(nw.col("a") == 4)
        .then(40)
        .otherwise(50)
        .alias("result")
    )

    expected = {"result": [10, 20, 30, 40, 50]}
    assert_equal_data(result, expected)


def test_when_chain_no_otherwise(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3, 4]}))

    result = df.select(
        nw.when(nw.col("a") == 1).then(10).when(nw.col("a") == 2).then(20).alias("result")
    )

    expected = {"result": [10, 20, None, None]}
    assert_equal_data(result, expected)


def test_when_chain_first_match_wins(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3, 5, 10]}))

    result = df.select(
        nw.when(nw.col("a") < 5)
        .then(100)  # Matches 1, 2, 3
        .when(nw.col("a") < 3)
        .then(200)  # Would match 1, 2 but first condition already matched
        .when(nw.col("a") < 2)
        .then(300)  # Would match 1 but first condition already matched
        .otherwise(400)
        .alias("result")
    )

    expected = {"result": [100, 100, 100, 400, 400]}
    assert_equal_data(result, expected)


def test_when_chain_all_conditions_false(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3]}))

    result = df.select(
        nw.when(nw.col("a") > 10)
        .then(10)
        .when(nw.col("a") > 20)
        .then(20)
        .when(nw.col("a") > 30)
        .then(30)
        .otherwise(999)
        .alias("result")
    )

    expected = {"result": [999, 999, 999]}
    assert_equal_data(result, expected)


def test_when_chain_mixed_value_types(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3], "b": [10, 20, 30]}))

    result = df.select(
        nw.when(nw.col("a") == 1)
        .then(nw.col("b"))  # Column reference
        .when(nw.col("a") == 2)
        .then(100)  # Scalar literal
        .when(nw.col("a") == 3)
        .then(nw.col("b") * 2)  # Expression
        .otherwise(0)
        .alias("result")
    )

    expected = {"result": [10, 100, 60]}
    assert_equal_data(result, expected)


def test_when_chain_complex_conditions(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3, 4, 5], "b": [10, 20, 30, 40, 50]}))

    result = df.select(
        nw.when((nw.col("a") == 1) & (nw.col("b") == 10))
        .then(111)
        .when((nw.col("a") > 2) & (nw.col("b") < 40))
        .then(222)
        .when(nw.col("a") + nw.col("b") > 50)
        .then(333)
        .otherwise(444)
        .alias("result")
    )

    expected = {"result": [111, 444, 222, 444, 333]}
    assert_equal_data(result, expected)


def test_when_chain_with_nulls(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, None, 3, None, 5]}))

    result = df.select(
        nw.when(nw.col("a") == 1)
        .then(10)
        .when(nw.col("a") == 3)
        .then(30)
        .when(nw.col("a") == 5)
        .then(50)
        .otherwise(99)
        .alias("result")
    )

    # Null values don't match any condition, so they get otherwise value
    expected = {"result": [10, 99, 30, 99, 50]}
    assert_equal_data(result, expected)


def test_when_chain_expression_conditions(constructor: Constructor) -> None:
    if uses_pyarrow_backend(constructor) or "pyarrow" in str(constructor):
        reason = (
            "PyArrow backend doesn't support modulo operator. "
            "This is a pre-existing PyArrow arithmetic limitation."
        )
        pytest.skip(reason=reason)

    df = nw.from_native(constructor({"a": [1, 2, 3, 4, 5]}))

    result = df.select(
        nw.when(nw.col("a") % 2 == 0)
        .then(1000)  # Even numbers
        .when(nw.col("a") % 3 == 0)
        .then(2000)  # Odd multiples of 3
        .otherwise(3000)
        .alias("result")
    )

    # 1: odd, not div by 3 -> 3000
    # 2: even -> 1000
    # 3: odd, div by 3 -> 2000 (first match)
    # 4: even -> 1000
    # 5: odd, not div by 3 -> 3000
    # Note: 3 is divisible by both 2 (false) and 3 (true)
    expected = {"result": [3000, 1000, 2000, 1000, 3000]}
    assert_equal_data(result, expected)


def test_when_chain_value_types_string(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3], "b": ["x", "y", "z"]}))

    result = df.select(
        nw.when(nw.col("a") == 1)
        .then(nw.lit("first"))
        .when(nw.col("a") == 2)
        .then(nw.col("b"))
        .otherwise(nw.lit("other"))
        .alias("result")
    )

    expected = {"result": ["first", "y", "other"]}
    assert_equal_data(result, expected)


def test_when_chain_value_types_float(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3], "b": [1.1, 2.2, 3.3]}))

    result = df.select(
        nw.when(nw.col("a") == 1)
        .then(10.5)
        .when(nw.col("a") == 2)
        .then(nw.col("b"))
        .when(nw.col("a") == 3)
        .then(nw.col("b") * 2.0)
        .otherwise(0.0)
        .alias("result")
    )

    expected = {"result": [10.5, 2.2, 6.6]}
    assert_equal_data(result, expected)


def test_when_chain_value_types_mixed(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3], "b": [10, 20, 30]}))

    result = df.select(
        nw.when(nw.col("a") == 1)
        .then(nw.col("b"))
        .when(nw.col("a") == 2)
        .then(100)
        .when(nw.col("a") == 3)
        .then(nw.col("b") * 2)
        .otherwise(0)
        .alias("result")
    )

    expected = {"result": [10, 100, 60]}
    assert_equal_data(result, expected)


def test_when_chain_with_columns(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3], "b": [10, 20, 30]}))

    result = df.with_columns(
        nw.when(nw.col("a") == 1)
        .then(100)
        .when(nw.col("a") == 2)
        .then(200)
        .otherwise(300)
        .alias("c")
    )

    expected = {"a": [1, 2, 3], "b": [10, 20, 30], "c": [100, 200, 300]}
    assert_equal_data(result, expected)


def test_when_chain_multiple_columns(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3]}))

    result = df.select(
        nw.when(nw.col("a") == 1)
        .then(10)
        .when(nw.col("a") == 2)
        .then(20)
        .otherwise(30)
        .alias("x"),
        nw.when(nw.col("a") < 2)
        .then(100)
        .when(nw.col("a") < 3)
        .then(200)
        .otherwise(300)
        .alias("y"),
    )

    expected = {"x": [10, 20, 30], "y": [100, 200, 300]}
    assert_equal_data(result, expected)


def test_when_chain_backward_compatible_single_condition(
    constructor: Constructor,
) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3]}))

    # This is the existing API that must continue to work
    result = df.select(nw.when(nw.col("a") == 1).then(10).otherwise(20).alias("result"))

    expected = {"result": [10, 20, 20]}
    assert_equal_data(result, expected)


def test_when_chain_two_conditions_no_otherwise(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3, 4, 5]}))

    result = df.select(
        nw.when(nw.col("a") == 1).then(10).when(nw.col("a") == 2).then(20).alias("result")
    )

    expected = {"result": [10, 20, None, None, None]}
    assert_equal_data(result, expected)


def test_when_chain_conditions_on_multiple_columns(constructor: Constructor) -> None:
    df = nw.from_native(
        constructor({"a": [1, 2, 3, 4], "b": [10, 20, 30, 40], "c": [5, 15, 25, 35]})
    )

    result = df.select(
        nw.when(nw.col("a") == 1)
        .then(nw.col("b"))
        .when(nw.col("c") > 20)
        .then(nw.col("c"))
        .otherwise(999)
        .alias("result")
    )

    # Row 0: a==1 -> b=10
    # Row 1: a!=1, c=15 not >20 -> 999
    # Row 2: a!=1, c=25 >20 -> 25
    # Row 3: a!=1, c=35 >20 -> 35
    expected = {"result": [10, 999, 25, 35]}
    assert_equal_data(result, expected)


def test_when_chain_boolean_column_condition(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [True, False, True, False], "b": [1, 2, 3, 4]}))

    result = df.select(
        nw.when(nw.col("a"))
        .then(100)
        .when(nw.col("b") > 2)
        .then(200)
        .otherwise(300)
        .alias("result")
    )

    # Row 0: a=True -> 100
    # Row 1: a=False, b=2 not >2 -> 300
    # Row 2: a=True -> 100
    # Row 3: a=False, b=4 >2 -> 200
    expected = {"result": [100, 300, 100, 200]}
    assert_equal_data(result, expected)


def test_when_chain_branching_does_not_mutate(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3, 4]}))

    base_expr = nw.when(nw.col("a") == 1).then(10)
    derived_expr = base_expr.when(nw.col("a") > 2).then(nw.col("a") * 10).otherwise(0)

    result = df.select(base_expr.alias("base"), derived_expr.alias("derived"))
    expected = {"base": [10, None, None, None], "derived": [10, 0, 30, 40]}
    assert_equal_data(result, expected)


def test_when_then_composes_with_expr_operations(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 2, 3, 4]}))

    result = df.select(
        (
            nw.when(nw.col("a") < 3)
            .then(nw.col("a"))
            .when(nw.col("a") == 3)
            .then(10)
            .otherwise(0)
            * 2
        ).alias("result")
    )

    expected = {"result": [2, 4, 20, 0]}
    assert_equal_data(result, expected)



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Literal

import pytest

import narwhals as nw
from tests.utils import (
    PANDAS_VERSION,
    Constructor,
    assert_equal_data,
    pyspark_session,
    sqlframe_session,
)

pytest.importorskip("polars")
pytest.importorskip("pyarrow")
import polars as pl

if TYPE_CHECKING:
    from collections.abc import Mapping
    from pathlib import Path
    from types import ModuleType

    from typing_extensions import TypeAlias

    from narwhals._typing import EagerAllowed, _LazyOnly, _SparkLike
    from narwhals.typing import FileSource

    Factory: TypeAlias = pytest.TempPathFactory

IOSourceKind: TypeAlias = Literal["str", "Path", "PathLike"]

data: Mapping[str, Any] = {"a": [1, 2, 3], "b": [4.5, 6.7, 8.9], "z": ["x", "y", "w"]}
skipif_pandas_lt_1_5 = pytest.mark.skipif(
    PANDAS_VERSION < (1, 5), reason="too old for pyarrow"
)
lazy_core_backend = pytest.mark.parametrize("backend", ["duckdb", "ibis", "sqlframe"])
spark_like_backend = pytest.mark.parametrize("backend", ["pyspark", "sqlframe"])


class MockPathLike:
    def __init__(self, path: Path) -> None:
        self._super_secret: Path = path

    def __fspath__(self) -> str:
        return self._super_secret.__fspath__()


def _into_file_source(source: Path, which: IOSourceKind, /) -> FileSource:
    mapping: Mapping[IOSourceKind, FileSource] = {
        "str": str(source),
        "Path": source,
        "PathLike": MockPathLike(source),
    }
    return mapping[which]


def _path(factory: Factory, name: str, /) -> Path:
    # NOTE: Generates a path on windows that contains `\\n` and `\\t`
    # See https://github.com/narwhals-dev/narwhals/issues/3422
    tmp_dir = factory.mktemp("newline")
    sub_dir = tmp_dir / "tab"
    sub_dir.mkdir(exist_ok=True)
    return sub_dir / name


@pytest.fixture(scope="module", params=["str", "Path", "PathLike"])
def csv_path(tmp_path_factory: Factory, request: pytest.FixtureRequest) -> FileSource:
    fp = _path(tmp_path_factory, "file.csv")
    pl.DataFrame(data).write_csv(fp)
    return _into_file_source(fp, request.param)


@pytest.fixture(scope="module", params=["str", "Path", "PathLike"])
def csv_path_sep(tmp_path_factory: Factory, request: pytest.FixtureRequest) -> FileSource:
    fp = _path(tmp_path_factory, "file.csv")
    pl.DataFrame(data).write_csv(fp, separator="|")
    return _into_file_source(fp, request.param)


@pytest.fixture(scope="module", params=["str", "Path", "PathLike"])
def parquet_path(tmp_path_factory: Factory, request: pytest.FixtureRequest) -> FileSource:
    fp = _path(tmp_path_factory, "file.parquet")
    pl.DataFrame(data).write_parquet(fp)
    return _into_file_source(fp, request.param)


def assert_equal_eager(result: nw.DataFrame[Any]) -> None:
    assert_equal_data(result, data)
    assert isinstance(result, nw.DataFrame)


def assert_equal_lazy(result: nw.LazyFrame[Any]) -> None:
    assert_equal_data(result, data)
    assert isinstance(result, nw.LazyFrame)


def native_namespace(cb: Constructor, /) -> ModuleType:
    return nw.get_native_namespace(nw.from_native(cb(data)))  # type: ignore[no-any-return]


def test_read_csv(
    csv_path: FileSource, csv_path_sep: FileSource, eager_backend: EagerAllowed
) -> None:
    assert_equal_eager(nw.read_csv(csv_path, backend=eager_backend))
    assert_equal_eager(nw.read_csv(csv_path_sep, backend=eager_backend, separator="|"))


@skipif_pandas_lt_1_5
def test_read_csv_kwargs(csv_path: FileSource) -> None:
    pytest.importorskip("pandas")
    pytest.importorskip("pyarrow")
    import pandas as pd
    from pyarrow import csv

    assert_equal_eager(nw.read_csv(csv_path, backend=pd, engine="pyarrow"))
    assert_equal_eager(
        nw.read_csv(
            csv_path, backend="pyarrow", parse_options=csv.ParseOptions(delimiter=",")
        )
    )


@lazy_core_backend
def test_read_csv_raise_with_lazy(backend: _LazyOnly) -> None:
    pytest.importorskip(backend)
    with pytest.raises(ValueError, match="Expected eager backend, found"):
        nw.read_csv("unused.csv", backend=backend)  # type: ignore[arg-type]


def test_scan_csv(
    csv_path: FileSource, csv_path_sep: FileSource, constructor: Constructor
) -> None:
    kwargs: dict[str, Any]
    if "sqlframe" in str(constructor):
        kwargs = {"session": sqlframe_session(), "inferSchema": True, "header": True}
    elif "pyspark" in str(constructor):
        kwargs = {"session": pyspark_session(), "inferSchema": True, "header": True}
    else:
        kwargs = {}
    backend = native_namespace(constructor)
    assert_equal_lazy(nw.scan_csv(csv_path, backend=backend, **kwargs))
    assert_equal_lazy(nw.scan_csv(csv_path_sep, backend=backend, separator="|", **kwargs))


@skipif_pandas_lt_1_5
def test_scan_csv_kwargs(csv_path: FileSource) -> None:
    pytest.importorskip("pandas")
    pytest.importorskip("pyarrow")
    import pandas as pd

    assert_equal_data(nw.scan_csv(csv_path, backend=pd, engine="pyarrow"), data)


@skipif_pandas_lt_1_5
def test_read_parquet(parquet_path: FileSource, eager_backend: EagerAllowed) -> None:
    assert_equal_eager(nw.read_parquet(parquet_path, backend=eager_backend))


@skipif_pandas_lt_1_5
def test_read_parquet_kwargs(parquet_path: FileSource) -> None:
    pytest.importorskip("pandas")
    pytest.importorskip("pyarrow")
    import pandas as pd

    assert_equal_eager(nw.read_parquet(parquet_path, backend=pd, engine="pyarrow"))


@lazy_core_backend
def test_read_parquet_raise_with_lazy(backend: _LazyOnly) -> None:
    pytest.importorskip(backend)
    with pytest.raises(ValueError, match="Expected eager backend, found"):
        nw.read_parquet("unused.parquet", backend=backend)  # type: ignore[arg-type]


@skipif_pandas_lt_1_5
def test_scan_parquet(parquet_path: FileSource, constructor: Constructor) -> None:
    kwargs: dict[str, Any]
    if "sqlframe" in str(constructor):
        kwargs = {"session": sqlframe_session(), "inferSchema": True}
    elif "pyspark" in str(constructor):
        kwargs = {"session": pyspark_session(), "inferSchema": True, "header": True}
    else:
        kwargs = {}
    backend = native_namespace(constructor)
    assert_equal_lazy(nw.scan_parquet(parquet_path, backend=backend, **kwargs))


@skipif_pandas_lt_1_5
def test_scan_parquet_kwargs(parquet_path: FileSource) -> None:
    pytest.importorskip("pandas")
    pytest.importorskip("pyarrow")
    import pandas as pd

    assert_equal_lazy(nw.scan_parquet(parquet_path, backend=pd, engine="pyarrow"))


@spark_like_backend
@pytest.mark.parametrize("scan_method", ["scan_csv", "scan_parquet"])
def test_scan_fail_spark_like_without_session(
    backend: _SparkLike, scan_method: str
) -> None:
    pytest.importorskip(backend)
    pattern = re.compile(r"spark.+backend.+require.+session", re.IGNORECASE)
    with pytest.raises(ValueError, match=pattern):
        getattr(nw, scan_method)("unused.csv", backend=backend)


@pytest.mark.parametrize("csv_path", ["str"], indirect=True)
def test_read_csv_raise_sep_multiple_lazy(csv_path: FileSource) -> None:
    pytest.importorskip("duckdb")
    pytest.importorskip("pandas")
    pytest.importorskip("pyarrow")
    pytest.importorskip("sqlframe")
    import duckdb
    import pandas as pd
    import pyarrow as pa
    import sqlframe
    from pyarrow import csv
    from sqlframe.duckdb import DuckDBSession

    msg = "do not match:"
    with pytest.raises(TypeError, match=msg):
        nw.read_csv(
            csv_path,
            backend=pa,
            separator="|",
            parse_options=csv.ParseOptions(delimiter=";"),
        )
    with pytest.raises(TypeError, match=msg):
        nw.scan_csv(
            csv_path,
            backend=pa,
            separator="|",
            parse_options=csv.ParseOptions(delimiter=";"),
        )
    with pytest.raises(TypeError, match=msg):
        nw.read_csv(csv_path, backend=pd, separator="|", sep=";")
    with pytest.raises(TypeError, match=msg):
        nw.scan_csv(csv_path, backend=pd, separator="|", sep=";")
    with pytest.raises(TypeError, match=msg):
        nw.scan_csv(csv_path, backend=duckdb, separator="|", delimiter=";")
    with pytest.raises(TypeError, match=msg):
        nw.scan_csv(csv_path, backend=duckdb, separator="|", delim=";")
    with pytest.raises(TypeError, match=msg):
        nw.scan_csv(
            csv_path,
            backend=sqlframe,
            separator="|",
            sep=";",
            session=DuckDBSession(),
            inferSchema=True,
        )
    with pytest.raises(TypeError, match=msg):
        nw.scan_csv(
            csv_path,
            backend=sqlframe,
            separator="|",
            delimiter=";",
            session=DuckDBSession(),
            inferSchema=True,
        )



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Literal

import hypothesis.strategies as st
import pytest
from hypothesis import given

import narwhals as nw
from tests.utils import (
    PANDAS_VERSION,
    POLARS_VERSION,
    Constructor,
    ConstructorEager,
    assert_equal_data,
    is_pyarrow_windows_no_tzdata,
    time_unit_compat,
)

if TYPE_CHECKING:
    from narwhals.typing import IntoSeriesT

data = {
    "a": [datetime(2021, 3, 1, 12, 34, 56, 49000), datetime(2020, 1, 2, 2, 4, 14, 715000)]
}


@pytest.mark.parametrize(
    ("original_time_unit", "time_unit", "expected"),
    [
        ("ns", "ns", [978307200000000000, None, 978480000000000000]),
        ("ns", "us", [978307200000000, None, 978480000000000]),
        ("ns", "ms", [978307200000, None, 978480000000]),
        ("us", "ns", [978307200000000000, None, 978480000000000000]),
        ("us", "us", [978307200000000, None, 978480000000000]),
        ("us", "ms", [978307200000, None, 978480000000]),
        ("ms", "ns", [978307200000000000, None, 978480000000000000]),
        ("ms", "us", [978307200000000, None, 978480000000000]),
        ("ms", "ms", [978307200000, None, 978480000000]),
        ("s", "ns", [978307200000000000, None, 978480000000000000]),
        ("s", "us", [978307200000000, None, 978480000000000]),
        ("s", "ms", [978307200000, None, 978480000000]),
    ],
)
def test_timestamp_datetimes(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    original_time_unit: Literal["us", "ns", "ms", "s"],
    time_unit: Literal["ns", "us", "ms"],
    expected: list[int | None],
) -> None:
    if "dask" in str(constructor):
        pytest.skip(reason="https://github.com/narwhals-dev/narwhals/issues/2808")
    if any(x in str(constructor) for x in ("duckdb", "pyspark", "ibis")):
        request.applymarker(
            pytest.mark.xfail(reason="Backend timestamp conversion not yet implemented")
        )
    if original_time_unit == "s" and "polars" in str(constructor):
        pytest.skip("Second precision not supported in Polars")

    if "pandas_pyarrow" in str(constructor) and PANDAS_VERSION < (
        2,
        2,
    ):  # pragma: no cover
        pytest.skip("Requires pandas >= 2.2 for reliable pyarrow-backed timestamps")
    datetimes = {"a": [datetime(2001, 1, 1), None, datetime(2001, 1, 3)]}
    df = nw.from_native(constructor(datetimes))
    dtype = nw.Datetime(time_unit_compat(original_time_unit, request))
    result = df.select(nw.col("a").cast(dtype).dt.timestamp(time_unit))
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(
    ("original_time_unit", "time_unit", "expected"),
    [
        ("ns", "ns", [978307200000000000, None, 978480000000000000]),
        ("ns", "us", [978307200000000, None, 978480000000000]),
        ("ns", "ms", [978307200000, None, 978480000000]),
        ("us", "ns", [978307200000000000, None, 978480000000000000]),
        ("us", "us", [978307200000000, None, 978480000000000]),
        ("us", "ms", [978307200000, None, 978480000000]),
        ("ms", "ns", [978307200000000000, None, 978480000000000000]),
        ("ms", "us", [978307200000000, None, 978480000000000]),
        ("ms", "ms", [978307200000, None, 978480000000]),
        ("s", "ns", [978307200000000000, None, 978480000000000000]),
        ("s", "us", [978307200000000, None, 978480000000000]),
        ("s", "ms", [978307200000, None, 978480000000]),
    ],
)
def test_timestamp_datetimes_tz_aware(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    original_time_unit: Literal["us", "ns", "ms", "s"],
    time_unit: Literal["ns", "us", "ms"],
    expected: list[int | None],
) -> None:
    if "dask" in str(constructor):
        pytest.skip(reason="https://github.com/narwhals-dev/narwhals/issues/2808")
    if any(x in str(constructor) for x in ("duckdb", "pyspark", "ibis")):
        request.applymarker(
            pytest.mark.xfail(reason="Backend timestamp conversion not yet implemented")
        )
    version_conditions = [
        (
            is_pyarrow_windows_no_tzdata(constructor),
            "Timezone database is not installed on Windows",
        ),
        (
            "pandas_pyarrow" in str(constructor) and PANDAS_VERSION < (2,),
            "Requires pandas >= 2.0 for pyarrow support",
        ),
        (
            "pandas_pyarrow" in str(constructor) and PANDAS_VERSION < (2, 2),
            "Requires pandas >= 2.2 for reliable timestamps",
        ),
        (
            "dask" in str(constructor) and PANDAS_VERSION < (2, 1),
            "Requires pandas >= 2.1 for dask support",
        ),
    ]

    for condition, reason in version_conditions:
        if condition:
            pytest.skip(reason)  # pragma: no cover

    if original_time_unit == "s" and "polars" in str(constructor):
        pytest.skip()
    datetimes = {"a": [datetime(2001, 1, 1), None, datetime(2001, 1, 3)]}
    df = nw.from_native(constructor(datetimes))
    dtype = nw.Datetime(time_unit_compat(original_time_unit, request))
    result = df.select(
        nw.col("a")
        .cast(dtype)
        .dt.replace_time_zone("UTC")
        .dt.convert_time_zone("Asia/Kathmandu")
        .dt.timestamp(time_unit)
    )
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(
    ("time_unit", "expected"),
    [
        ("ns", [978307200000000000, None, 978480000000000000]),
        ("us", [978307200000000, None, 978480000000000]),
        ("ms", [978307200000, None, 978480000000]),
    ],
)
def test_timestamp_dates(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    time_unit: Literal["ns", "us", "ms"],
    expected: list[int | None],
) -> None:
    if "dask" in str(constructor):
        pytest.skip(reason="https://github.com/narwhals-dev/narwhals/issues/2808")
    if any(x in str(constructor) for x in ("duckdb", "pyspark", "ibis")):
        request.applymarker(
            pytest.mark.xfail(reason="Backend timestamp conversion not yet implemented")
        )
    unsupported_backends = (
        "pandas_constructor",
        "pandas_nullable_constructor",
        "cudf",
        "modin_constructor",
    )
    if any(x in str(constructor) for x in unsupported_backends):
        pytest.skip("Backend does not support date type")

    dates = {"a": [datetime(2001, 1, 1), None, datetime(2001, 1, 3)]}
    if "dask" in str(constructor):
        df = nw.from_native(
            constructor(dates).astype({"a": "timestamp[ns][pyarrow]"})  # type: ignore[union-attr]
        )
    else:
        df = nw.from_native(constructor(dates))
    result = df.select(nw.col("a").dt.date().dt.timestamp(time_unit))
    assert_equal_data(result, {"a": expected})


def test_timestamp_invalid_date(
    request: pytest.FixtureRequest, constructor: Constructor
) -> None:
    if "dask" in str(constructor):
        pytest.skip(reason="https://github.com/narwhals-dev/narwhals/issues/2808")
    if any(x in str(constructor) for x in ("duckdb", "pyspark", "ibis")):
        request.applymarker(
            pytest.mark.xfail(reason="Backend timestamp conversion not yet implemented")
        )
    if "polars" in str(constructor):
        request.applymarker(
            pytest.mark.xfail(
                reason="Invalid date handling not yet implemented in Polars"
            )
        )
    data_str = {"a": ["x", "y", None]}
    data_num = {"a": [1, 2, None]}
    df_str = nw.from_native(constructor(data_str))
    df_num = nw.from_native(constructor(data_num))
    msg = "Input should be either of Date or Datetime type"
    with pytest.raises(TypeError, match=msg):
        df_str.select(nw.col("a").dt.timestamp())
    with pytest.raises(TypeError, match=msg):
        df_num.select(nw.col("a").dt.timestamp())


def test_timestamp_invalid_unit_expr(constructor: Constructor) -> None:
    time_unit_invalid = "i"
    msg = (
        "invalid `time_unit`"
        f"\n\nExpected one of {{'ns', 'us', 'ms'}}, got {time_unit_invalid!r}."
    )
    with pytest.raises(ValueError, match=msg):
        nw.from_native(constructor(data)).select(
            nw.col("a").dt.timestamp(time_unit_invalid)  # type: ignore[arg-type]
        )


def test_timestamp_invalid_unit_series(constructor_eager: ConstructorEager) -> None:
    time_unit_invalid = "i"
    msg = (
        "invalid `time_unit`"
        f"\n\nExpected one of {{'ns', 'us', 'ms'}}, got {time_unit_invalid!r}."
    )
    with pytest.raises(ValueError, match=msg):
        nw.from_native(constructor_eager(data))["a"].dt.timestamp(time_unit_invalid)  # type: ignore[arg-type]


@given(
    inputs=st.datetimes(min_value=datetime(1960, 1, 1), max_value=datetime(1980, 1, 1)),
    time_unit=st.sampled_from(["ms", "us", "ns"]),
    # We keep 'ms' out for now due to an upstream bug: https://github.com/pola-rs/polars/issues/19309
    starting_time_unit=st.sampled_from(["us", "ns"]),
)
@pytest.mark.skipif(PANDAS_VERSION < (2, 2), reason="bug in old pandas")
@pytest.mark.skipif(POLARS_VERSION < (0, 20, 7), reason="bug in old Polars")
@pytest.mark.slow
def test_timestamp_hypothesis(
    inputs: datetime,
    time_unit: Literal["ms", "us", "ns"],
    starting_time_unit: Literal["ms", "us", "ns"],
) -> None:
    pytest.importorskip("pandas")
    pytest.importorskip("pyarrow")
    import pandas as pd
    import polars as pl
    import pyarrow as pa

    @nw.narwhalify
    def func(s: nw.Series[IntoSeriesT]) -> nw.Series[IntoSeriesT]:
        return s.dt.timestamp(time_unit)

    result_pl = func(pl.Series([inputs], dtype=pl.Datetime(starting_time_unit)))
    result_pd = func(pd.Series([inputs], dtype=f"datetime64[{starting_time_unit}]"))
    result_pdpa = func(
        pd.Series([inputs], dtype=f"timestamp[{starting_time_unit}][pyarrow]")
    )
    result_pa = func(pa.chunked_array([[inputs]], type=pa.timestamp(starting_time_unit)))
    assert result_pl[0] == result_pd[0]
    assert result_pl[0] == result_pdpa[0]
    assert result_pl[0] == result_pa[0].as_py()



from __future__ import annotations

from typing import TYPE_CHECKING

import pyarrow as pa

from narwhals._arrow.utils import ArrowSeriesNamespace
from narwhals._compliant.any_namespace import CatNamespace

if TYPE_CHECKING:
    from narwhals._arrow.series import ArrowSeries
    from narwhals._arrow.typing import Incomplete


class ArrowSeriesCatNamespace(ArrowSeriesNamespace, CatNamespace["ArrowSeries"]):
    def get_categories(self) -> ArrowSeries:
        # NOTE: Should be `list[pa.DictionaryArray]`, but `DictionaryArray` has no attributes
        chunks: Incomplete = self.native.chunks
        return self.with_native(pa.concat_arrays(x.dictionary for x in chunks).unique())



from __future__ import annotations

from datetime import timedelta

import hypothesis.strategies as st
import pytest
from hypothesis import given

import narwhals as nw
from tests.utils import PANDAS_VERSION

pytest.importorskip("pandas")
import pandas as pd


@given(
    timedeltas=st.timedeltas(
        min_value=-timedelta(days=5, minutes=70, seconds=10),
        max_value=timedelta(days=3, minutes=90, seconds=60),
    )
)
@pytest.mark.skipif(PANDAS_VERSION < (2, 2, 0), reason="pyarrow dtype not available")
@pytest.mark.slow
def test_total_minutes(timedeltas: timedelta) -> None:
    pytest.importorskip("pyarrow")
    result_pd = nw.from_native(
        pd.Series([timedeltas]), series_only=True
    ).dt.total_minutes()[0]
    result_pdns = nw.from_native(
        pd.Series([timedeltas]).dt.as_unit("ns"), series_only=True
    ).dt.total_minutes()[0]
    result_pda = nw.from_native(
        pd.Series([timedeltas]).convert_dtypes(dtype_backend="pyarrow"), series_only=True
    ).dt.total_minutes()[0]
    result_pdn = nw.from_native(
        pd.Series([timedeltas]).convert_dtypes(dtype_backend="numpy_nullable"),
        series_only=True,
    ).dt.total_minutes()[0]
    assert result_pda == result_pd
    assert result_pdn == result_pd
    assert result_pdns == result_pd


@given(
    timedeltas=st.timedeltas(
        min_value=-timedelta(days=5, minutes=70, seconds=10),
        max_value=timedelta(days=3, minutes=90, seconds=60),
    )
)
@pytest.mark.skipif(PANDAS_VERSION < (2, 2, 0), reason="pyarrow dtype not available")
@pytest.mark.slow
def test_total_minutes_polars(timedeltas: timedelta) -> None:
    pytest.importorskip("polars")
    pytest.importorskip("pyarrow")
    import polars as pl

    result_pd = nw.from_native(
        pd.Series([timedeltas]), series_only=True
    ).dt.total_minutes()[0]
    result_pl = nw.from_native(
        pl.Series([timedeltas]), series_only=True
    ).dt.total_minutes()[0]
    assert result_pl == result_pd



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(line_item_ds: FrameT, orders_ds: FrameT) -> FrameT:
    var1 = "MAIL"
    var2 = "SHIP"
    var3 = datetime(1994, 1, 1)
    var4 = datetime(1995, 1, 1)

    return (
        orders_ds.join(line_item_ds, left_on="o_orderkey", right_on="l_orderkey")
        .filter(nw.col("l_shipmode").is_in([var1, var2]))
        .filter(nw.col("l_commitdate") < nw.col("l_receiptdate"))
        .filter(nw.col("l_shipdate") < nw.col("l_commitdate"))
        .filter(nw.col("l_receiptdate").is_between(var3, var4, closed="left"))
        .with_columns(
            nw.when(nw.col("o_orderpriority").is_in(["1-URGENT", "2-HIGH"]))
            .then(1)
            .otherwise(0)
            .alias("high_line_count"),
            nw.when(~nw.col("o_orderpriority").is_in(["1-URGENT", "2-HIGH"]))
            .then(1)
            .otherwise(0)
            .alias("low_line_count"),
        )
        .group_by("l_shipmode")
        .agg(nw.col("high_line_count").sum(), nw.col("low_line_count").sum())
        .sort("l_shipmode")
    )



from __future__ import annotations

import operator
from functools import reduce
from itertools import chain
from typing import TYPE_CHECKING, Any

from narwhals._expression_parsing import (
    combine_alias_output_names,
    combine_evaluate_output_names,
    evaluate_output_names_and_aliases,
)
from narwhals._spark_like.dataframe import SparkLikeLazyFrame
from narwhals._spark_like.expr import SparkLikeExpr
from narwhals._spark_like.selectors import SparkLikeSelectorNamespace
from narwhals._spark_like.utils import (
    import_functions,
    import_native_dtypes,
    narwhals_to_native_dtype,
    true_divide,
)
from narwhals._sql.namespace import SQLNamespace
from narwhals._utils import zip_strict

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from sqlframe.base.column import Column

    from narwhals._compliant.window import WindowInputs
    from narwhals._spark_like.dataframe import SQLFrameDataFrame  # noqa: F401
    from narwhals._utils import Implementation, Version
    from narwhals.typing import ConcatMethod, CorrelationMethod, IntoDType, PythonLiteral

# Adjust slight SQL vs PySpark differences
FUNCTION_REMAPPINGS = {
    "starts_with": "startswith",
    "ends_with": "endswith",
    "trim": "btrim",
    "str_split": "split",
    "regexp_matches": "regexp",
}


class SparkLikeNamespace(
    SQLNamespace[SparkLikeLazyFrame, SparkLikeExpr, "SQLFrameDataFrame", "Column"]
):
    def __init__(self, *, version: Version, implementation: Implementation) -> None:
        self._version = version
        self._implementation = implementation

    @property
    def selectors(self) -> SparkLikeSelectorNamespace:
        return SparkLikeSelectorNamespace.from_namespace(self)

    @property
    def _expr(self) -> type[SparkLikeExpr]:
        return SparkLikeExpr

    @property
    def _lazyframe(self) -> type[SparkLikeLazyFrame]:
        return SparkLikeLazyFrame

    @property
    def _F(self):  # type: ignore[no-untyped-def] # noqa: ANN202
        if TYPE_CHECKING:
            from sqlframe.base import functions

            return functions
        return import_functions(self._implementation)

    @property
    def _native_dtypes(self):  # type: ignore[no-untyped-def] # noqa: ANN202
        if TYPE_CHECKING:
            from sqlframe.base import types

            return types
        return import_native_dtypes(self._implementation)

    def _function(self, name: str, *args: Column | PythonLiteral) -> Column:
        return getattr(self._F, FUNCTION_REMAPPINGS.get(name, name))(*args)

    def _lit(self, value: Any) -> Column:
        return self._F.lit(value)

    def _when(
        self, condition: Column, value: Column, otherwise: Column | None = None
    ) -> Column:
        if otherwise is None:
            return self._F.when(condition, value)
        return self._F.when(condition, value).otherwise(otherwise)

    def _coalesce(self, *exprs: Column) -> Column:
        return self._F.coalesce(*exprs)

    def lit(self, value: PythonLiteral, dtype: IntoDType | None) -> SparkLikeExpr:
        def func(df: SparkLikeLazyFrame) -> list[Column]:
            F = df._F

            if isinstance(value, (list, tuple)):
                lit_values = [F.lit(v) for v in value]
                column = F.lit(F.array(lit_values))
            elif isinstance(value, dict):
                if not value:
                    impl = self._implementation
                    if impl.is_sqlframe():
                        msg = f"Cannot create an empty struct type for {self._implementation} backend"
                        raise NotImplementedError(msg)

                    column = (  # pragma: no cover
                        F.from_json(F.lit("{}"), df._native_dtypes.StructType())
                        if impl._backend_version() < (4, 0)
                        else F.struct([])
                    )
                else:
                    lit_values = [F.lit(v).alias(k) for k, v in value.items()]
                    column = F.struct(*lit_values)
            else:
                column = F.lit(value)

            if dtype:
                native_dtype = narwhals_to_native_dtype(
                    dtype, self._version, df._native_dtypes, df.native.sparkSession
                )
                column = column.cast(native_dtype)

            return [column]

        def window_func(
            df: SparkLikeLazyFrame, _window_inputs: WindowInputs[Column]
        ) -> list[Column]:
            return func(df)

        return self._expr(
            func,
            window_func,
            evaluate_output_names=lambda _df: ["literal"],
            alias_output_names=None,
            version=self._version,
            implementation=self._implementation,
        )

    def len(self) -> SparkLikeExpr:
        def func(df: SparkLikeLazyFrame) -> list[Column]:
            return [df._F.count("*")]

        return self._expr(
            func,
            evaluate_output_names=lambda _df: ["len"],
            alias_output_names=None,
            version=self._version,
            implementation=self._implementation,
        )

    def mean_horizontal(self, *exprs: SparkLikeExpr) -> SparkLikeExpr:
        def func(cols: Iterable[Column]) -> Column:
            cols = tuple(cols)
            F = exprs[0]._F
            numerator = reduce(
                operator.add, (self._F.coalesce(col, self._F.lit(0)) for col in cols)
            )
            denominator = reduce(
                operator.add,
                (col.isNotNull().cast(self._native_dtypes.IntegerType()) for col in cols),
            )
            return true_divide(F, numerator, denominator)

        return self._expr._from_elementwise_horizontal_op(func, *exprs)

    def concat(
        self, items: Iterable[SparkLikeLazyFrame], *, how: ConcatMethod
    ) -> SparkLikeLazyFrame:
        dfs = [item._native_frame for item in items]
        if how == "vertical":
            cols_0 = dfs[0].columns
            for i, df in enumerate(dfs[1:], start=1):
                cols_current = df.columns
                if not ((len(cols_current) == len(cols_0)) and (cols_current == cols_0)):
                    msg = (
                        "unable to vstack, column names don't match:\n"
                        f"   - dataframe 0: {cols_0}\n"
                        f"   - dataframe {i}: {cols_current}\n"
                    )
                    raise TypeError(msg)

            return SparkLikeLazyFrame(
                native_dataframe=reduce(lambda x, y: x.union(y), dfs),
                version=self._version,
                implementation=self._implementation,
            )

        if how == "diagonal":
            return SparkLikeLazyFrame(
                native_dataframe=reduce(
                    lambda x, y: x.unionByName(y, allowMissingColumns=True), dfs
                ),
                version=self._version,
                implementation=self._implementation,
            )
        raise NotImplementedError

    def concat_str(
        self, *exprs: SparkLikeExpr, separator: str, ignore_nulls: bool
    ) -> SparkLikeExpr:
        def func(df: SparkLikeLazyFrame) -> list[Column]:
            F = self._F
            cols = tuple(chain.from_iterable(e(df) for e in exprs))
            result = F.concat_ws(separator, *cols)

            if not ignore_nulls:
                null_mask = reduce(operator.or_, (F.isnull(s) for s in cols))
                result = F.when(~null_mask, result).otherwise(F.lit(None))

            return [result]

        return self._expr(
            call=func,
            evaluate_output_names=combine_evaluate_output_names(*exprs),
            alias_output_names=combine_alias_output_names(*exprs),
            version=self._version,
            implementation=self._implementation,
        )

    def corr(
        self, a: SparkLikeExpr, b: SparkLikeExpr, *, method: CorrelationMethod
    ) -> SparkLikeExpr:
        if method != "pearson":
            msg = "Only 'pearson' correlation is supported for Spark."
            raise NotImplementedError(msg)

        def func(df: SparkLikeLazyFrame) -> list[Column]:
            F = self._F
            a_ = df._evaluate_single_output_expr(a)
            b_ = df._evaluate_single_output_expr(b)
            return [F.corr(a_, b_)]

        return self._expr(
            call=func,
            evaluate_output_names=combine_evaluate_output_names(a, b),
            alias_output_names=combine_alias_output_names(a, b),
            version=self._version,
            implementation=self._implementation,
        )

    def struct(self, *exprs: SparkLikeExpr) -> SparkLikeExpr:
        version = self._version

        def func(df: SparkLikeLazyFrame) -> list[Column]:
            F = self._F
            names_to_cols: Mapping[str, Column] = {
                alias: native_expr
                for expr in exprs
                for native_expr, _, alias in zip_strict(
                    expr(df), *evaluate_output_names_and_aliases(expr, df, [])
                )
            }
            aliased = (col.alias(name) for name, col in names_to_cols.items())
            return [F.struct(*aliased)]

        return self._expr(
            call=func,
            evaluate_output_names=combine_evaluate_output_names(*exprs),
            alias_output_names=combine_alias_output_names(*exprs),
            version=version,
            implementation=self._implementation,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.plugins import Plugin


def test_plugin() -> None:
    pytest.importorskip("test_plugin")
    df_native = {"a": [1, 1, 2], "b": [4, 5, 6]}
    lf = nw.from_native(df_native)  # type: ignore[call-overload]
    assert isinstance(lf, nw.LazyFrame)
    assert lf.columns == ["a", "b"]


def test_not_implemented() -> None:
    pytest.importorskip("test_plugin")
    df_native = {"a": [1, 1, 2], "b": [4, 5, 6]}
    lf = nw.from_native(df_native)  # type: ignore[call-overload]
    with pytest.raises(
        NotImplementedError, match="is not implemented for: 'DictLazyFrame'"
    ):
        lf.select(nw.col("a").ewm_mean())


def test_typing() -> None:
    pytest.importorskip("test_plugin")
    import test_plugin

    _plugin: Plugin = test_plugin



"""Ensuring backports/extensions to new `typing` features are understood correctly."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Literal

import pytest

from narwhals._typing_compat import assert_never


def test_assert_never() -> None:
    pattern = re.compile(
        r"expected.+unreachable.+got.+'a'.+report.+issue.+github.+narwhals",
        re.DOTALL | re.IGNORECASE,
    )
    some: Literal["a"] = "a"
    if some != "a":
        assigned = "b"
        assert_never(assigned)
    else:
        assigned = some
    if not TYPE_CHECKING:
        # NOTE: Trying to avoid the assert influencing narrowing
        assert assigned == "a"
    with pytest.raises(AssertionError, match=pattern):
        assert_never(assigned)  # type: ignore[arg-type]



from __future__ import annotations

import inspect
import string
import sys

# ruff: noqa: N806
from collections import OrderedDict, deque
from inspect import isfunction, ismethoddescriptor
from pathlib import Path
from types import MethodType, ModuleType
from typing import TYPE_CHECKING, Any

import polars as pl

import narwhals as nw

if TYPE_CHECKING:
    from collections.abc import Iterator

LOWERCASE = tuple(string.ascii_lowercase)

if sys.version_info >= (3, 13):

    def _is_public_method_or_property(obj: Any) -> bool:
        return (
            isfunction(obj)
            or (isinstance(obj, (MethodType, property)) or ismethoddescriptor(obj))
        ) and obj.__name__.startswith(LOWERCASE)
else:

    def _is_public_method_or_property(obj: Any) -> bool:
        return (
            (isfunction(obj) or (isinstance(obj, MethodType) or ismethoddescriptor(obj)))
            and obj.__name__.startswith(LOWERCASE)
        ) or (isinstance(obj, property) and obj.fget.__name__.startswith(LOWERCASE))


def iter_api_reference_names(tp: type[Any]) -> Iterator[str]:
    for name, _ in inspect.getmembers(tp, _is_public_method_or_property):
        yield name


def iter_api_reference_names_dtypes(module: ModuleType) -> Iterator[str]:
    base = module.DType
    # NOTE: Special case, only non-dtype
    field_dtype = module.Field
    for name, _ in inspect.getmembers(
        module,
        lambda x: isinstance(x, type) and (issubclass(x, base) or x is field_dtype),
    ):
        yield name


def read_documented_members(source: str | Path) -> list[str]:
    MEMBERS_START = "members:\n"
    MEMBERS_PREFIX = "        - "
    DUNDER_PREFIX = "__"

    with Path(source).open(encoding="utf-8") as fd:
        lines = deque(fd.readlines())
    head = lines.popleft()
    while not head.endswith(MEMBERS_START):
        head = lines.popleft()
    while not head.startswith(MEMBERS_PREFIX):
        head = lines.pop()
    lines.append(head)
    all_members = (line.removeprefix(MEMBERS_PREFIX).strip() for line in lines)
    return [m for m in all_members if not m.startswith(DUNDER_PREFIX)]


ret = 0

NAMESPACES = {"dt", "str", "cat", "name", "list", "struct"}
EXPR_ONLY_METHODS = {"over", "map_batches"}
SERIES_ONLY_METHODS = {
    "arg_max",
    "arg_min",
    "arg_true",
    "dtype",
    "from_iterable",
    "from_numpy",
    "gather_every",
    "implementation",
    "is_empty",
    "is_sorted",
    "head",
    "hist",
    "item",
    "name",
    "rename",
    "sample",
    "scatter",
    "shape",
    "sort",
    "tail",
    "to_arrow",
    "to_dummies",
    "to_frame",
    "to_list",
    "to_native",
    "to_numpy",
    "to_pandas",
    "to_polars",
    "value_counts",
    "zip_with",
    "__iter__",
    "__contains__",
}
DIR_API_REF = Path("docs/api-reference")

files = {fp.stem for fp in Path("narwhals").iterdir()}

# Top level functions
top_level_functions = [
    i for i in nw.__all__ if not i[0].isupper() and i[0] != "_" and i not in files
]

documented = read_documented_members(DIR_API_REF / "narwhals.md")
if missing := set(top_level_functions).difference(documented).difference({"annotations"}):
    print("top-level functions: not documented")
    print(missing)
    ret = 1
if extra := set(documented).difference(top_level_functions):
    print("top-level functions: outdated")
    print(extra)
    ret = 1

# DataFrame methods
dataframe_methods = list(iter_api_reference_names(nw.DataFrame))
documented = read_documented_members(DIR_API_REF / "dataframe.md")
if missing := set(dataframe_methods).difference(documented):
    print("DataFrame: not documented")
    print(missing)
    ret = 1
if extra := set(documented).difference(dataframe_methods):
    print("DataFrame: outdated")
    print(extra)
    ret = 1

# LazyFrame methods
lazyframe_methods = list(iter_api_reference_names(nw.LazyFrame))
documented = read_documented_members(DIR_API_REF / "lazyframe.md")
if missing := set(lazyframe_methods).difference(documented):
    print("LazyFrame: not documented")
    print(missing)
    ret = 1
if extra := set(documented).difference(lazyframe_methods):
    print("LazyFrame: outdated")
    print(extra)
    ret = 1

# Series methods
series_methods = list(iter_api_reference_names(nw.Series))
documented = read_documented_members(DIR_API_REF / "series.md")
if missing := set(series_methods).difference(documented).difference(NAMESPACES):
    print("Series: not documented")
    print(missing)
    ret = 1
if extra := set(documented).difference(series_methods):
    print("Series: outdated")
    print(extra)
    ret = 1

# Series.{cat, dt, list, str} methods
for namespace in NAMESPACES.difference({"name"}):
    series_ns_methods = [
        i
        for i in dir(getattr(nw.from_native(pl.Series(), series_only=True), namespace))
        if not i[0].isupper() and i[0] != "_"
    ]
    documented = read_documented_members(DIR_API_REF / f"series_{namespace}.md")
    if missing := set(series_ns_methods).difference(documented):
        print(f"Series.{namespace}: not documented")
        print(missing)
        ret = 1
    if extra := set(documented).difference(series_ns_methods):
        print(f"Series.{namespace}: outdated")
        print(extra)
        ret = 1

# Expr methods
expr_methods = list(iter_api_reference_names(nw.Expr))
documented = read_documented_members(DIR_API_REF / "expr.md")
if missing := set(expr_methods).difference(documented).difference(NAMESPACES):
    print("Expr: not documented")
    print(missing)
    ret = 1
if extra := set(documented).difference(expr_methods):
    print("Expr: outdated")
    print(extra)
    ret = 1

# Expr.{cat, dt, list, name, str} methods
for namespace in NAMESPACES:
    expr_ns_methods = [
        i
        for i in dir(getattr(nw.col("a"), namespace))
        if not i[0].isupper() and i[0] != "_"
    ]
    documented = read_documented_members(DIR_API_REF / f"expr_{namespace}.md")
    if missing := set(expr_ns_methods).difference(documented):
        print(f"Expr.{namespace}: not documented")
        print(missing)
        ret = 1
    if extra := set(documented).difference(expr_ns_methods):
        print(f"Expr.{namespace}: outdated")
        print(extra)
        ret = 1

# DTypes
dtypes = list(iter_api_reference_names_dtypes(nw.dtypes))
documented = read_documented_members(DIR_API_REF / "dtypes.md")
if missing := set(dtypes).difference(documented):
    print("Dtype: not documented")
    print(missing)
    ret = 1
if extra := set(documented).difference(dtypes):
    print("Dtype: outdated")
    print(extra)
    ret = 1

# Schema
schema_methods = list(iter_api_reference_names(nw.Schema))
documented = read_documented_members(DIR_API_REF / "schema.md")
if (
    missing := set(schema_methods)
    .difference(documented)
    .difference(iter_api_reference_names(OrderedDict))
):
    print("Schema: not documented")
    print(missing)
    ret = 1
if extra := set(documented).difference(schema_methods):
    print("Schema: outdated")
    print(extra)
    ret = 1

# Check Expr vs Series
if missing := set(expr_methods).difference(series_methods).difference(EXPR_ONLY_METHODS):
    print("In Expr but not in Series")
    print(missing)
    ret = 1
if extra := set(series_methods).difference(expr_methods).difference(SERIES_ONLY_METHODS):
    print("In Series but not in Expr")
    print(extra)
    ret = 1

# Check Expr vs Series internal methods
for namespace in NAMESPACES.difference({"name"}):
    expr_internal = [
        i
        for i in dir(getattr(nw.col("a"), namespace))
        if not i[0].isupper() and i[0] != "_"
    ]
    series_internal = [
        i
        for i in dir(getattr(nw.from_native(pl.Series(), series_only=True), namespace))
        if not i[0].isupper() and i[0] != "_"
    ]
    if missing := set(expr_internal).difference(series_internal):
        print(f"In Expr.{namespace} but not in Series.{namespace}")
        print(missing)
        ret = 1
    if extra := set(series_internal).difference(expr_internal):
        print(f"In Series.{namespace} but not in Expr.{namespace}")
        print(extra)
        ret = 1

sys.exit(ret)



from __future__ import annotations

from typing import TYPE_CHECKING

from narwhals._constants import (
    MS_PER_MINUTE,
    MS_PER_SECOND,
    NS_PER_SECOND,
    SECONDS_PER_MINUTE,
    US_PER_MINUTE,
    US_PER_SECOND,
)
from narwhals._duckdb.utils import UNITS_DICT, F, fetch_rel_time_zone, lit, sql_expression
from narwhals._duration import Interval
from narwhals._sql.expr_dt import SQLExprDateTimeNamesSpace
from narwhals._utils import not_implemented, requires

if TYPE_CHECKING:
    from collections.abc import Sequence

    from duckdb import Expression

    from narwhals._duckdb.dataframe import DuckDBLazyFrame
    from narwhals._duckdb.expr import DuckDBExpr


class DuckDBExprDateTimeNamespace(SQLExprDateTimeNamesSpace["DuckDBExpr"]):
    def millisecond(self) -> DuckDBExpr:
        return self.compliant._with_elementwise(
            lambda expr: F("millisecond", expr) - F("second", expr) * lit(MS_PER_SECOND)
        )

    def microsecond(self) -> DuckDBExpr:
        return self.compliant._with_elementwise(
            lambda expr: F("microsecond", expr) - F("second", expr) * lit(US_PER_SECOND)
        )

    def nanosecond(self) -> DuckDBExpr:
        return self.compliant._with_elementwise(
            lambda expr: F("nanosecond", expr) - F("second", expr) * lit(NS_PER_SECOND)
        )

    def to_string(self, format: str) -> DuckDBExpr:
        return self.compliant._with_elementwise(
            lambda expr: F("strftime", expr, lit(format))
        )

    def weekday(self) -> DuckDBExpr:
        return self.compliant._with_elementwise(lambda expr: F("isodow", expr))

    def date(self) -> DuckDBExpr:
        return self.compliant._with_elementwise(lambda expr: expr.cast("date"))

    def total_minutes(self) -> DuckDBExpr:
        return self.compliant._with_elementwise(
            lambda expr: F("datepart", lit("minute"), expr)
        )

    def total_seconds(self) -> DuckDBExpr:
        return self.compliant._with_elementwise(
            lambda expr: (
                lit(SECONDS_PER_MINUTE) * F("datepart", lit("minute"), expr)
                + F("datepart", lit("second"), expr)
            )
        )

    def total_milliseconds(self) -> DuckDBExpr:
        return self.compliant._with_elementwise(
            lambda expr: (
                lit(MS_PER_MINUTE) * F("datepart", lit("minute"), expr)
                + F("datepart", lit("millisecond"), expr)
            )
        )

    def total_microseconds(self) -> DuckDBExpr:
        return self.compliant._with_elementwise(
            lambda expr: (
                lit(US_PER_MINUTE) * F("datepart", lit("minute"), expr)
                + F("datepart", lit("microsecond"), expr)
            )
        )

    def truncate(self, every: str) -> DuckDBExpr:
        interval = Interval.parse(every)
        multiple, unit = interval.multiple, interval.unit
        if multiple != 1:
            # https://github.com/duckdb/duckdb/issues/17554
            msg = f"Only multiple 1 is currently supported for DuckDB.\nGot {multiple!s}."
            raise ValueError(msg)
        if unit == "ns":
            msg = "Truncating to nanoseconds is not yet supported for DuckDB."
            raise NotImplementedError(msg)
        format = lit(UNITS_DICT[unit])

        def _truncate(expr: Expression) -> Expression:
            return F("date_trunc", format, expr)

        return self.compliant._with_elementwise(_truncate)

    @requires.backend_version((1, 3))
    def offset_by(self, by: str) -> DuckDBExpr:
        interval = Interval.parse_no_constraints(by)
        format = lit(f"{interval.multiple!s} {UNITS_DICT[interval.unit]}")

        def _offset_by(expr: Expression) -> Expression:
            return expr + sql_expression(f"interval {format}")

        return self.compliant._with_elementwise(_offset_by)

    def _no_op_time_zone(self, time_zone: str) -> DuckDBExpr:
        def func(df: DuckDBLazyFrame) -> Sequence[Expression]:
            native_series_list = self.compliant(df)
            if (conn_time_zone := fetch_rel_time_zone(df.native)) != time_zone:
                msg = (
                    "DuckDB stores the time zone in the connection, rather than in the "
                    f"data type, so changing the timezone to anything other than {conn_time_zone} "
                    " (the current connection time zone) is not supported."
                )
                raise NotImplementedError(msg)
            return native_series_list

        return self.compliant.__class__(
            func,
            evaluate_output_names=self.compliant._evaluate_output_names,
            alias_output_names=self.compliant._alias_output_names,
            version=self.compliant._version,
        )

    def convert_time_zone(self, time_zone: str) -> DuckDBExpr:
        return self._no_op_time_zone(time_zone)

    def replace_time_zone(self, time_zone: str | None) -> DuckDBExpr:
        if time_zone is None:
            return self.compliant._with_elementwise(lambda expr: expr.cast("timestamp"))
        return self._no_op_time_zone(time_zone)

    total_nanoseconds = not_implemented()
    timestamp = not_implemented()



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any, Protocol, overload

from narwhals._compliant.typing import (
    CompliantExprT,
    CompliantExprT_co,
    CompliantFrameT,
    CompliantLazyFrameT,
    DepthTrackingExprT,
    EagerDataFrameT,
    EagerExprT,
    EagerSeriesT_co,
    LazyExprT,
    NativeFrameT,
    NativeSeriesT,
)
from narwhals._utils import (
    exclude_column_names,
    get_column_names,
    passthrough_column_names,
)
from narwhals.dependencies import is_numpy_array_2d

if TYPE_CHECKING:
    from collections.abc import Collection, Iterable, Iterator, KeysView, Sequence

    from typing_extensions import TypeAlias, TypeIs

    from narwhals._compliant.selectors import CompliantSelectorNamespace
    from narwhals._utils import Implementation, Version
    from narwhals.typing import (
        ConcatMethod,
        CorrelationMethod,
        Into1DArray,
        IntoDType,
        IntoSchema,
        NonNestedLiteral,
        _2DArray,
    )

    Incomplete: TypeAlias = Any

__all__ = [
    "CompliantNamespace",
    "DepthTrackingNamespace",
    "EagerNamespace",
    "LazyNamespace",
]


class CompliantNamespace(Protocol[CompliantFrameT, CompliantExprT]):
    # NOTE: `narwhals`
    _implementation: Implementation
    _version: Version

    @property
    def _expr(self) -> type[CompliantExprT]: ...
    # NOTE: `polars`
    def all(self) -> CompliantExprT:
        return self._expr.from_column_names(get_column_names, context=self)

    def col(self, *names: str) -> CompliantExprT:
        return self._expr.from_column_names(passthrough_column_names(names), context=self)

    def exclude(self, *names: str) -> CompliantExprT:
        return self._expr.from_column_names(
            partial(exclude_column_names, names=names), context=self
        )

    def nth(self, indices: Sequence[int]) -> CompliantExprT:
        return self._expr.from_column_indices(*indices, context=self)

    def corr(
        self, a: CompliantExprT, b: CompliantExprT, *, method: CorrelationMethod
    ) -> CompliantExprT: ...
    def len(self) -> CompliantExprT: ...
    def lit(self, value: NonNestedLiteral, dtype: IntoDType | None) -> CompliantExprT: ...
    def all_horizontal(
        self, *exprs: CompliantExprT, ignore_nulls: bool
    ) -> CompliantExprT: ...
    def any_horizontal(
        self, *exprs: CompliantExprT, ignore_nulls: bool
    ) -> CompliantExprT: ...
    def sum_horizontal(self, *exprs: CompliantExprT) -> CompliantExprT: ...
    def mean_horizontal(self, *exprs: CompliantExprT) -> CompliantExprT: ...
    def min_horizontal(self, *exprs: CompliantExprT) -> CompliantExprT: ...
    def max_horizontal(self, *exprs: CompliantExprT) -> CompliantExprT: ...
    def concat(
        self, items: Iterable[CompliantFrameT], *, how: ConcatMethod
    ) -> CompliantFrameT: ...
    def concat_str(
        self, *exprs: CompliantExprT, separator: str, ignore_nulls: bool
    ) -> CompliantExprT: ...
    @property
    def selectors(self) -> CompliantSelectorNamespace[Any, Any]: ...
    def struct(self, *exprs: CompliantExprT) -> CompliantExprT: ...
    def coalesce(self, *exprs: CompliantExprT) -> CompliantExprT: ...
    # NOTE: typing this accurately requires 2x more `TypeVar`s
    def from_native(self, data: Any, /) -> Any: ...
    def is_native(self, obj: Any, /) -> TypeIs[Any]:
        """Return `True` if `obj` can be passed to `from_native`."""
        ...


class AlignDiagonal(Protocol[CompliantFrameT, CompliantExprT_co]):
    """Mixin to help support `"diagonal*"` concatenation."""

    def lit(
        self, value: NonNestedLiteral, dtype: IntoDType | None
    ) -> CompliantExprT_co: ...
    def align_diagonal(
        self, frames: Collection[CompliantFrameT], /
    ) -> Sequence[CompliantFrameT]:
        """Prepare frames with differing schemas for vertical concatenation.

        Adapted from [`convert_diagonal_concat`].

        [`convert_diagonal_concat`]: https://github.com/pola-rs/polars/blob/c2412600210a21143835c9dfcb0a9182f462b619/crates/polars-plan/src/plans/conversion/dsl_to_ir/concat.rs#L10-L68
        """
        schemas = [frame.collect_schema() for frame in frames]
        # 1 - Take the first schema, collecting any fields from the remaining that introduce new names
        # Subtle difference from `dict |= dict |= ...` as we preserve the first `DType`, rather than the last.
        it_schemas = iter(schemas)
        union = dict(next(it_schemas))
        seen = union.keys()
        for schema in it_schemas:
            union.update((nm, dtype) for nm, dtype in schema.items() if nm not in seen)
        return self._align_diagonal(frames, schemas, union)

    def _align_diagonal(
        self,
        frames: Iterable[CompliantFrameT],
        schemas: Iterable[IntoSchema],
        union_schema: IntoSchema,
    ) -> Sequence[CompliantFrameT]:
        union_names = union_schema.keys()
        # Lazily populate null expressions as needed, shared across frames
        null_exprs: dict[str, CompliantExprT_co] = {}

        def iter_missing_exprs(missing: Iterable[str]) -> Iterator[CompliantExprT_co]:
            nonlocal null_exprs
            for name in missing:
                if (expr := null_exprs.get(name)) is None:
                    dtype = union_schema[name]
                    expr = null_exprs[name] = self.lit(None, dtype).alias(name)
                yield expr

        def align(df: CompliantFrameT, columns: KeysView[str]) -> CompliantFrameT:
            if missing := union_names - columns:
                df = df.with_columns(*iter_missing_exprs(missing))
            # Even if all fields are present, we always reorder the columns to match between frames.
            return df.simple_select(*union_names)

        return [align(frame, schema.keys()) for frame, schema in zip(frames, schemas)]


class DepthTrackingNamespace(
    CompliantNamespace[CompliantFrameT, DepthTrackingExprT],
    Protocol[CompliantFrameT, DepthTrackingExprT],
):
    def all(self) -> DepthTrackingExprT:
        return self._expr.from_column_names(get_column_names, context=self)

    def col(self, *names: str) -> DepthTrackingExprT:
        return self._expr.from_column_names(passthrough_column_names(names), context=self)

    def exclude(self, *names: str) -> DepthTrackingExprT:
        return self._expr.from_column_names(
            partial(exclude_column_names, names=names), context=self
        )


class LazyNamespace(
    CompliantNamespace[CompliantLazyFrameT, LazyExprT],
    Protocol[CompliantLazyFrameT, LazyExprT, NativeFrameT],
):
    @property
    def _backend_version(self) -> tuple[int, ...]:
        return self._implementation._backend_version()

    @property
    def _lazyframe(self) -> type[CompliantLazyFrameT]: ...
    def is_native(self, obj: Any, /) -> TypeIs[NativeFrameT]:
        return self._lazyframe._is_native(obj)

    def from_native(self, data: NativeFrameT | Any, /) -> CompliantLazyFrameT:
        if self._lazyframe._is_native(data):
            return self._lazyframe.from_native(data, context=self)
        msg = f"Unsupported type: {type(data).__name__!r}"  # pragma: no cover
        raise TypeError(msg)


class EagerNamespace(
    DepthTrackingNamespace[EagerDataFrameT, EagerExprT],
    Protocol[EagerDataFrameT, EagerSeriesT_co, EagerExprT, NativeFrameT, NativeSeriesT],
):
    @property
    def _backend_version(self) -> tuple[int, ...]:
        return self._implementation._backend_version()

    @property
    def _dataframe(self) -> type[EagerDataFrameT]: ...
    @property
    def _series(self) -> type[EagerSeriesT_co]: ...
    def _if_then_else(
        self,
        when: NativeSeriesT,
        then: NativeSeriesT,
        otherwise: NativeSeriesT | None = None,
    ) -> NativeSeriesT: ...
    def when_then(
        self, predicate: EagerExprT, then: EagerExprT, otherwise: EagerExprT | None = None
    ) -> EagerExprT:
        def func(df: EagerDataFrameT) -> Sequence[EagerSeriesT_co]:
            predicate_s = df._evaluate_single_output_expr(predicate)
            align = predicate_s._align_full_broadcast

            then_s = df._evaluate_single_output_expr(then)
            if otherwise is None:
                predicate_s, then_s = align(predicate_s, then_s)
                result = self._if_then_else(predicate_s.native, then_s.native)

            if otherwise is None:
                predicate_s, then_s = align(predicate_s, then_s)
                result = self._if_then_else(predicate_s.native, then_s.native)
            else:
                otherwise_s = df._evaluate_single_output_expr(otherwise)
                predicate_s, then_s, otherwise_s = align(predicate_s, then_s, otherwise_s)
                result = self._if_then_else(
                    predicate_s.native, then_s.native, otherwise_s.native
                )
            return [then_s._with_native(result)]

        return self._expr._from_callable(
            func=func,
            evaluate_output_names=getattr(
                then, "_evaluate_output_names", lambda _df: ["literal"]
            ),
            alias_output_names=getattr(then, "_alias_output_names", None),
            context=predicate,
        )

    def is_native(self, obj: Any, /) -> TypeIs[NativeFrameT | NativeSeriesT]:
        return self._dataframe._is_native(obj) or self._series._is_native(obj)

    @overload
    def from_native(self, data: NativeFrameT, /) -> EagerDataFrameT: ...
    @overload
    def from_native(self, data: NativeSeriesT, /) -> EagerSeriesT_co: ...
    def from_native(
        self, data: NativeFrameT | NativeSeriesT | Any, /
    ) -> EagerDataFrameT | EagerSeriesT_co:
        if self._dataframe._is_native(data):
            return self._dataframe.from_native(data, context=self)
        if self._series._is_native(data):
            return self._series.from_native(data, context=self)
        msg = f"Unsupported type: {type(data).__name__!r}"
        raise TypeError(msg)

    @overload
    def from_numpy(self, data: Into1DArray, /, schema: None = ...) -> EagerSeriesT_co: ...

    @overload
    def from_numpy(
        self, data: _2DArray, /, schema: IntoSchema | Sequence[str] | None
    ) -> EagerDataFrameT: ...

    def from_numpy(
        self,
        data: Into1DArray | _2DArray,
        /,
        schema: IntoSchema | Sequence[str] | None = None,
    ) -> EagerDataFrameT | EagerSeriesT_co:
        if is_numpy_array_2d(data):
            return self._dataframe.from_numpy(data, schema=schema, context=self)
        return self._series.from_numpy(data, context=self)

    def _concat_diagonal(self, dfs: Sequence[NativeFrameT], /) -> NativeFrameT: ...
    def _concat_horizontal(
        self, dfs: Sequence[NativeFrameT | Any], /
    ) -> NativeFrameT: ...
    def _concat_vertical(self, dfs: Sequence[NativeFrameT], /) -> NativeFrameT: ...
    def concat(
        self, items: Iterable[EagerDataFrameT], *, how: ConcatMethod
    ) -> EagerDataFrameT:
        dfs = [item.native for item in items]
        if how == "horizontal":
            native = self._concat_horizontal(dfs)
        elif how == "vertical":
            native = self._concat_vertical(dfs)
        elif how == "diagonal":
            native = self._concat_diagonal(dfs)
        else:  # pragma: no cover
            raise NotImplementedError
        return self._dataframe.from_native(native, context=self)



from __future__ import annotations

import re
from contextlib import nullcontext as does_not_raise
from typing import Any

import pytest

import narwhals as nw
from narwhals.exceptions import ShapeError
from tests.utils import POLARS_VERSION, Constructor, ConstructorEager, assert_equal_data

data = {"a": [1, 1, 2, 2, 3], "b": [1, 2, 3, 3, 4]}
data_group = {
    "grp": ["g1", "g1", "g1", "g1", "g2", "g2", "g2"],
    "vals_unimodal": [2, 1, 1, 3, 3, 3, 4],
    "vals_multimodal_num": [1, 1, 2, 2, 3, 3, 4],
    "vals_multimodal_str": ["foo", "foo", "bar", "bar", "baz", "baz", "foo"],
}


def test_mode_single_expr_keep_all(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager(data))
    result = df.select(nw.col("a").mode(keep="all")).sort("a")
    expected = {"a": [1, 2]}
    assert_equal_data(result, expected)


def test_mode_series_keep_all(constructor_eager: ConstructorEager) -> None:
    series = nw.from_native(constructor_eager(data), eager_only=True)["a"]
    result = series.mode(keep="all").sort()
    expected = {"a": [1, 2]}
    assert_equal_data({"a": result}, expected)


def test_mode_series_keep_any(constructor_eager: ConstructorEager) -> None:
    frame = nw.from_native(constructor_eager(data), eager_only=True)
    result_a = frame["a"].mode(keep="any")
    assert result_a in {1, 2}

    result_b = frame["b"].mode(keep="any")
    assert result_b == 3


def test_mode_different_lengths_keep_all(constructor_eager: ConstructorEager) -> None:
    if "polars" in str(constructor_eager) and POLARS_VERSION < (1, 10):
        pytest.skip()
    df = nw.from_native(constructor_eager(data))
    with pytest.raises(ShapeError):
        df.select(nw.col("a", "b").mode(keep="all"))


def test_mode_expr_keep_any(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(nw.col("a", "b").mode(keep="any")).lazy().collect()

    try:
        expected = {"a": [1], "b": [3]}
        assert_equal_data(result, expected)
    except AssertionError:  # pragma: no cover
        expected = {"a": [2], "b": [3]}
        assert_equal_data(result, expected)


def test_mode_expr_keep_all_lazy(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    impl = df.implementation
    not_implemented = {
        nw.Implementation.DUCKDB,
        nw.Implementation.IBIS,
        nw.Implementation.PYSPARK,
        nw.Implementation.PYSPARK_CONNECT,
        nw.Implementation.SQLFRAME,
    }
    msg = re.escape("`Expr.mode(keep='all')` is not implemented for backend")
    context = (
        pytest.raises(NotImplementedError, match=msg)
        if impl in not_implemented
        else does_not_raise()
    )

    with context:
        result = df.select(nw.col("a").mode(keep="all").sum())
        assert_equal_data(result, {"a": [3]})


def test_mode_group_by_unimodal(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    df = nw.from_native(constructor(data_group))
    impl = df.implementation

    if impl.is_dask() or impl.is_pyarrow():
        # Issue tracker:
        #   - Dask: https://github.com/narwhals-dev/narwhals/pull/3019#issuecomment-3216649862
        #   - PyArrow: https://github.com/apache/arrow/issues/20359
        request.applymarker(pytest.mark.xfail)

    result = (
        df.group_by("grp")
        .agg(nw.col("vals_unimodal").mode(keep="any"))
        .sort("grp")
        .lazy()
        .collect()
    )
    expected = {"grp": ["g1", "g2"], "vals_unimodal": [1, 3]}
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("mode_col", "expected_opt_1", "expected_opt_2"),
    [
        ("vals_multimodal_num", [1, 3], [2, 3]),
        ("vals_multimodal_str", ["foo", "baz"], ["bar", "baz"]),
    ],
)
def test_mode_group_by_multimodal(
    constructor: Constructor,
    request: pytest.FixtureRequest,
    mode_col: str,
    expected_opt_1: list[Any],
    expected_opt_2: list[Any],
) -> None:
    df = nw.from_native(constructor(data_group))
    impl = df.implementation

    if impl.is_dask() or impl.is_pyarrow():
        # Issue tracker:
        #   - Dask: https://github.com/narwhals-dev/narwhals/pull/3019#issuecomment-3216649862
        #   - PyArrow: https://github.com/apache/arrow/issues/20359
        request.applymarker(pytest.mark.xfail)

    result = (
        df.group_by("grp")
        .agg(nw.col(mode_col).mode(keep="any"))
        .sort("grp")
        .lazy()
        .collect()
    )
    try:
        expected = {"grp": ["g1", "g2"], mode_col: expected_opt_1}
        assert_equal_data(result, expected)
    except AssertionError:  # pragma: no cover
        expected = {"grp": ["g1", "g2"], mode_col: expected_opt_2}
        assert_equal_data(result, expected)


@pytest.mark.parametrize(
    "mode_expr",
    [
        nw.col("vals_unimodal", "vals_multimodal_num").mode(keep="any"),
        [
            nw.col("vals_unimodal").mode(keep="any"),
            nw.col("vals_multimodal_num").mode(keep="any"),
        ],
    ],
)
def test_mode_group_by_multiple_cols(
    constructor: Constructor,
    request: pytest.FixtureRequest,
    mode_expr: nw.Expr | list[nw.Expr],
) -> None:
    df = nw.from_native(constructor(data_group))
    impl = df.implementation

    if impl.is_dask() or impl.is_pyarrow():
        # Issue tracker:
        #   - Dask: https://github.com/narwhals-dev/narwhals/pull/3019#issuecomment-3216649862
        #   - PyArrow: https://github.com/apache/arrow/issues/20359
        request.applymarker(pytest.mark.xfail)

    result = df.group_by("grp").agg(mode_expr).sort("grp").lazy().collect()

    try:
        expected = {
            "grp": ["g1", "g2"],
            "vals_unimodal": [1, 3],
            "vals_multimodal_num": [1, 3],
        }
        assert_equal_data(result, expected)
    except AssertionError:  # pragma: no cover
        expected = {
            "grp": ["g1", "g2"],
            "vals_unimodal": [1, 3],
            "vals_multimodal_num": [2, 3],
        }
        assert_equal_data(result, expected)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, Literal, Protocol

from narwhals._compliant.any_namespace import (
    CatNamespace,
    DateTimeNamespace,
    ListNamespace,
    StringNamespace,
    StructNamespace,
)
from narwhals._compliant.column import CompliantColumn
from narwhals._compliant.typing import (
    CompliantSeriesT_co,
    EagerDataFrameAny,
    EagerSeriesT,
    EagerSeriesT_co,
    NativeSeriesT,
    NativeSeriesT_co,
)
from narwhals._translate import FromIterable, FromNative, NumpyConvertible, ToNarwhals
from narwhals._typing_compat import TypeVar, assert_never
from narwhals._utils import (
    _StoresCompliant,
    _StoresNative,
    is_compliant_series,
    is_sized_multi_index_selector,
    unstable,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence
    from types import ModuleType

    import pandas as pd
    import polars as pl
    import pyarrow as pa
    from typing_extensions import NotRequired, Self, TypedDict

    from narwhals._compliant.dataframe import CompliantDataFrame
    from narwhals._compliant.namespace import EagerNamespace
    from narwhals._utils import Implementation, Version, _LimitedContext
    from narwhals.dtypes import DType
    from narwhals.series import Series
    from narwhals.typing import (
        Into1DArray,
        IntoDType,
        MultiIndexSelector,
        PythonLiteral,
        RollingInterpolationMethod,
        SizedMultiIndexSelector,
        _1DArray,
        _SliceIndex,
    )

    class HistData(TypedDict, Generic[NativeSeriesT, "_CountsT_co"]):
        breakpoint: NotRequired[list[float] | _1DArray | list[Any]]
        count: NativeSeriesT | _1DArray | _CountsT_co | list[Any]


_CountsT_co = TypeVar("_CountsT_co", bound="Iterable[Any]", covariant=True)

__all__ = [
    "CompliantSeries",
    "EagerSeries",
    "EagerSeriesCatNamespace",
    "EagerSeriesDateTimeNamespace",
    "EagerSeriesHist",
    "EagerSeriesListNamespace",
    "EagerSeriesNamespace",
    "EagerSeriesStringNamespace",
    "EagerSeriesStructNamespace",
]


class CompliantSeries(
    NumpyConvertible["_1DArray", "Into1DArray"],
    FromIterable,
    FromNative[NativeSeriesT],
    ToNarwhals["Series[NativeSeriesT]"],
    CompliantColumn,
    Protocol[NativeSeriesT],
):
    # NOTE: `narwhals`
    _implementation: Implementation

    @property
    def native(self) -> NativeSeriesT: ...
    def __narwhals_series__(self) -> Self:
        return self

    def __native_namespace__(self) -> ModuleType: ...
    @classmethod
    def from_native(cls, data: NativeSeriesT, /, *, context: _LimitedContext) -> Self: ...
    def to_narwhals(self) -> Series[NativeSeriesT]:
        return self._version.series(self, level="full")

    def _with_native(self, series: Any) -> Self: ...
    def _with_version(self, version: Version) -> Self: ...

    # NOTE: `polars`
    @property
    def dtype(self) -> DType: ...
    @property
    def name(self) -> str: ...
    def __array__(self, dtype: Any, *, copy: bool | None) -> _1DArray: ...
    def __contains__(self, other: Any) -> bool: ...
    def __getitem__(self, item: MultiIndexSelector[Self]) -> Any: ...
    def __iter__(self) -> Iterator[Any]: ...
    def __len__(self) -> int:
        return len(self.native)

    @classmethod
    def from_numpy(cls, data: Into1DArray, /, *, context: _LimitedContext) -> Self: ...
    @classmethod
    def from_iterable(
        cls,
        data: Iterable[Any],
        /,
        *,
        context: _LimitedContext,
        name: str = "",
        dtype: IntoDType | None = None,
    ) -> Self: ...
    def __radd__(self, other: Any) -> Self: ...
    def __rand__(self, other: Any) -> Self: ...
    def __rmul__(self, other: Any) -> Self: ...
    def __ror__(self, other: Any) -> Self: ...
    def all(self) -> bool: ...
    def any(self) -> bool: ...
    def any_value(self, *, ignore_nulls: bool) -> PythonLiteral: ...
    def arg_max(self) -> int: ...
    def arg_min(self) -> int: ...
    def arg_true(self) -> Self: ...
    def count(self) -> int: ...
    def filter(self, predicate: Any) -> Self: ...
    def first(self) -> PythonLiteral: ...
    def last(self) -> PythonLiteral: ...
    def gather_every(self, n: int, offset: int) -> Self: ...
    def head(self, n: int) -> Self: ...
    def is_empty(self) -> bool:
        return self.len() == 0

    def is_sorted(self, *, descending: bool) -> bool: ...
    def item(self, index: int | None) -> Any: ...
    def kurtosis(self) -> float | None: ...
    def len(self) -> int: ...
    def max(self) -> Any: ...
    def mean(self) -> float: ...
    def median(self) -> float: ...
    def min(self) -> Any: ...
    def n_unique(self) -> int: ...
    def null_count(self) -> int: ...
    def quantile(
        self, quantile: float, interpolation: RollingInterpolationMethod
    ) -> float: ...
    def sample(
        self,
        n: int | None,
        *,
        fraction: float | None,
        with_replacement: bool,
        seed: int | None,
    ) -> Self: ...
    def scatter(self, indices: Self, values: Self) -> Self: ...
    def shift(self, n: int) -> Self: ...
    def skew(self) -> float | None: ...
    def sort(self, *, descending: bool, nulls_last: bool) -> Self: ...
    def std(self, *, ddof: int) -> float: ...
    def sum(self) -> float: ...
    def tail(self, n: int) -> Self: ...
    def to_arrow(self) -> pa.Array[Any]: ...
    def to_dummies(
        self, *, separator: str, drop_first: bool
    ) -> CompliantDataFrame[Self, Any, Any, Any]: ...
    def to_frame(self) -> CompliantDataFrame[Self, Any, Any, Any]: ...
    def to_list(self) -> list[Any]: ...
    def to_pandas(self) -> pd.Series[Any]: ...
    def to_polars(self) -> pl.Series: ...
    def unique(self, *, maintain_order: bool = False) -> Self: ...
    def value_counts(
        self, *, sort: bool, parallel: bool, name: str | None, normalize: bool
    ) -> CompliantDataFrame[Self, Any, Any, Any]: ...
    def var(self, *, ddof: int) -> float: ...
    def zip_with(self, mask: Any, other: Any) -> Self: ...

    # NOTE: *Technically* `polars`
    @unstable
    def hist_from_bins(
        self, bins: list[float], *, include_breakpoint: bool
    ) -> CompliantDataFrame[Self, Any, Any, Any]:
        """`Series.hist(bins=..., bin_count=None)`."""
        ...

    @unstable
    def hist_from_bin_count(
        self, bin_count: int, *, include_breakpoint: bool
    ) -> CompliantDataFrame[Self, Any, Any, Any]:
        """`Series.hist(bins=None, bin_count=...)`."""
        ...


class EagerSeries(CompliantSeries[NativeSeriesT], Protocol[NativeSeriesT]):
    _native_series: Any
    _implementation: Implementation
    _version: Version
    _broadcast: bool

    @property
    def _backend_version(self) -> tuple[int, ...]:
        return self._implementation._backend_version()

    @classmethod
    def _align_full_broadcast(cls, *series: Self) -> Sequence[Self]:
        """Ensure all of `series` have the same length (and index if `pandas`).

        Scalars get broadcasted to the full length of the longest Series.

        This is useful when you need to construct a full Series anyway, such as:

            DataFrame.select(...)

        It should not be used in binary operations, such as:

            nw.col("a") - nw.col("a").mean()

        because then it's more efficient to extract the right-hand-side's single element as a scalar.
        """
        ...

    def _from_scalar(self, value: Any) -> Self:
        return self.from_iterable([value], name=self.name, context=self)

    def _with_native(
        self, series: NativeSeriesT, *, preserve_broadcast: bool = False
    ) -> Self:
        """Return a new `CompliantSeries`, wrapping the native `series`.

        In cases when operations are known to not affect whether a result should
        be broadcast, we can pass `preserve_broadcast=True`.
        Set this with care - it should only be set for unary expressions which don't
        change length or order, such as `.alias` or `.fill_null`. If in doubt, don't
        set it, you probably don't need it.
        """
        ...

    def __narwhals_namespace__(
        self,
    ) -> EagerNamespace[Any, Self, Any, Any, NativeSeriesT]: ...

    def _gather(self, rows: SizedMultiIndexSelector[NativeSeriesT]) -> Self: ...
    def _gather_slice(self, rows: _SliceIndex | range) -> Self: ...
    def __getitem__(self, item: MultiIndexSelector[Self]) -> Self:
        if isinstance(item, (slice, range)):
            return self._gather_slice(item)
        if is_compliant_series(item):
            return self._gather(item.native)
        elif is_sized_multi_index_selector(item):  # noqa: RET505
            return self._gather(item)
        assert_never(item)

    @property
    def str(self) -> EagerSeriesStringNamespace[Self, NativeSeriesT]: ...
    @property
    def dt(self) -> EagerSeriesDateTimeNamespace[Self, NativeSeriesT]: ...
    @property
    def cat(self) -> EagerSeriesCatNamespace[Self, NativeSeriesT]: ...
    @property
    def list(self) -> EagerSeriesListNamespace[Self, NativeSeriesT]: ...
    @property
    def struct(self) -> EagerSeriesStructNamespace[Self, NativeSeriesT]: ...


class _SeriesNamespace(  # type: ignore[misc]
    _StoresCompliant[CompliantSeriesT_co],
    _StoresNative[NativeSeriesT_co],
    Protocol[CompliantSeriesT_co, NativeSeriesT_co],
):
    _compliant_series: CompliantSeriesT_co

    @property
    def compliant(self) -> CompliantSeriesT_co:
        return self._compliant_series

    @property
    def implementation(self) -> Implementation:
        return self.compliant._implementation

    @property
    def backend_version(self) -> tuple[int, ...]:
        return self.implementation._backend_version()

    @property
    def version(self) -> Version:
        return self.compliant._version

    @property
    def native(self) -> NativeSeriesT_co:
        return self._compliant_series.native  # type: ignore[no-any-return]

    def with_native(self, series: Any, /) -> CompliantSeriesT_co:
        return self.compliant._with_native(series)


class EagerSeriesNamespace(
    _SeriesNamespace[EagerSeriesT_co, NativeSeriesT_co],
    Generic[EagerSeriesT_co, NativeSeriesT_co],
):
    _compliant_series: EagerSeriesT_co

    def __init__(self, series: EagerSeriesT_co, /) -> None:
        self._compliant_series = series


class EagerSeriesCatNamespace(  # type: ignore[misc]
    _SeriesNamespace[EagerSeriesT_co, NativeSeriesT_co],
    CatNamespace[EagerSeriesT_co],
    Protocol[EagerSeriesT_co, NativeSeriesT_co],
): ...


class EagerSeriesDateTimeNamespace(  # type: ignore[misc]
    _SeriesNamespace[EagerSeriesT_co, NativeSeriesT_co],
    DateTimeNamespace[EagerSeriesT_co],
    Protocol[EagerSeriesT_co, NativeSeriesT_co],
): ...


class EagerSeriesListNamespace(  # pyright: ignore[reportInvalidTypeVarUse]
    _SeriesNamespace[EagerSeriesT, NativeSeriesT_co],
    ListNamespace[EagerSeriesT],
    Protocol[EagerSeriesT, NativeSeriesT_co],
): ...


class EagerSeriesStringNamespace(
    _SeriesNamespace[EagerSeriesT, NativeSeriesT_co],
    StringNamespace[EagerSeriesT],
    Protocol[EagerSeriesT, NativeSeriesT_co],
): ...


class EagerSeriesStructNamespace(  # type: ignore[misc]
    _SeriesNamespace[EagerSeriesT_co, NativeSeriesT_co],
    StructNamespace[EagerSeriesT_co],
    Protocol[EagerSeriesT_co, NativeSeriesT_co],
): ...


class EagerSeriesHist(Protocol[NativeSeriesT, _CountsT_co]):
    _series: EagerSeries[NativeSeriesT]
    _breakpoint: bool
    _data: HistData[NativeSeriesT, _CountsT_co]

    @property
    def native(self) -> NativeSeriesT:
        return self._series.native

    @classmethod
    def from_series(
        cls, series: EagerSeries[NativeSeriesT], *, include_breakpoint: bool
    ) -> Self:
        obj = cls.__new__(cls)
        obj._series = series
        obj._breakpoint = include_breakpoint
        return obj

    def to_frame(self) -> EagerDataFrameAny: ...
    def _linear_space(  # NOTE: Roughly `pl.linear_space`
        self,
        start: float,
        end: float,
        num_samples: int,
        *,
        closed: Literal["both", "none"] = "both",
    ) -> _1DArray: ...

    # NOTE: *Could* be handled at narwhals-level
    def is_empty_series(self) -> bool: ...

    # NOTE: **Should** be handled at narwhals-level
    def data_empty(self) -> HistData[NativeSeriesT, _CountsT_co]:
        return {"breakpoint": [], "count": []} if self._breakpoint else {"count": []}

    # NOTE: *Could* be handled at narwhals-level, **iff** we add `nw.repeat`, `nw.linear_space`
    # See https://github.com/narwhals-dev/narwhals/pull/2839#discussion_r2215630696
    def series_empty(
        self, arg: int | list[float], /
    ) -> HistData[NativeSeriesT, _CountsT_co]: ...

    def with_bins(self, bins: list[float], /) -> Self:
        if len(bins) <= 1:
            self._data = self.data_empty()
        elif self.is_empty_series():
            self._data = self.series_empty(bins)
        else:
            self._data = self._calculate_hist(bins)
        return self

    def with_bin_count(self, bin_count: int, /) -> Self:
        if bin_count == 0:
            self._data = self.data_empty()
        elif self.is_empty_series():
            self._data = self.series_empty(bin_count)
        else:
            self._data = self._calculate_hist(self._calculate_bins(bin_count))
        return self

    def _calculate_breakpoint(self, arg: int | list[float], /) -> list[float] | _1DArray:
        bins = self._linear_space(0, 1, arg + 1) if isinstance(arg, int) else arg
        return bins[1:]

    def _calculate_bins(self, bin_count: int) -> _1DArray: ...
    def _calculate_hist(
        self, bins: list[float] | _1DArray
    ) -> HistData[NativeSeriesT, _CountsT_co]: ...



from __future__ import annotations

import re
from contextlib import AbstractContextManager, nullcontext as does_not_raise
from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw
from narwhals.testing import assert_frame_equal
from narwhals.testing.asserts.frame import GUARANTEES_ROW_ORDER
from tests.utils import PANDAS_VERSION

if TYPE_CHECKING:
    from narwhals.typing import IntoSchema
    from tests.conftest import Data
    from tests.utils import Constructor, ConstructorEager


def _assertion_error(detail: str) -> pytest.RaisesExc:
    msg = f"DataFrames are different ({detail})"
    return pytest.raises(AssertionError, match=re.escape(msg))


def test_check_narwhals_objects(constructor: Constructor) -> None:
    """Test that a type error is raised if the input is not a Narwhals object."""
    frame = constructor({"a": [1, 2, 3]})
    msg = re.escape(
        "Expected `narwhals.DataFrame` or `narwhals.LazyFrame` instance, found"
    )
    with pytest.raises(TypeError, match=msg):
        assert_frame_equal(frame, frame)  # type: ignore[arg-type]


def test_implementation_mismatch() -> None:
    """Test that different implementations raise an error."""
    pytest.importorskip("pandas")
    pytest.importorskip("pyarrow")

    import pandas as pd
    import pyarrow as pa

    with _assertion_error("implementation mismatch"):
        assert_frame_equal(
            nw.from_native(pd.DataFrame({"a": [1]})),
            nw.from_native(pa.table({"a": [1]})),  # type: ignore[type-var] # pyright: ignore[reportArgumentType]
        )


def test_check_same_input_type(constructor_eager: ConstructorEager) -> None:
    """Test that left and right frames are either both eager or both lazy.

    NOTE: Use `constructor_eager` instead of `constructor` so that the roundtrip
        `.lazy().collect()` preserves the same implementation (and we raise the check after)
    """
    frame = nw.from_native(constructor_eager({"a": [1, 2, 3]}))

    msg = re.escape("inputs are different (unexpected input types)")
    with pytest.raises(AssertionError, match=msg):
        assert_frame_equal(frame.lazy(), frame.lazy().collect())


@pytest.mark.parametrize(
    ("left_schema", "right_schema", "check_dtypes", "check_column_order", "context"),
    [
        # Same order, same dtypes
        (
            {"a": nw.Int32(), "b": nw.Float32()},
            {"a": nw.Int32(), "b": nw.Float32()},
            True,
            True,
            does_not_raise(),
        ),
        # Same order, different dtypes
        (
            {"a": nw.Int32(), "b": nw.Float32()},
            {"a": nw.Int32(), "b": nw.Float64()},
            False,
            True,
            does_not_raise(),
        ),
        (
            {"a": nw.Int32(), "b": nw.Float32()},
            {"a": nw.Int32(), "b": nw.Float64()},
            True,
            True,
            _assertion_error("dtypes do not match"),
        ),
        # Different order, same dtype
        (
            {"a": nw.Int32(), "b": nw.Float32()},
            {"b": nw.Float32(), "a": nw.Int32()},
            True,
            False,
            does_not_raise(),
        ),
        (
            {"a": nw.Int32(), "b": nw.Float32()},
            {"b": nw.Float32(), "a": nw.Int32()},
            True,
            True,
            _assertion_error("columns are not in the same order"),
        ),
        # Different order, different dtype
        (
            {"a": nw.Int32(), "b": nw.Float32()},
            {"b": nw.Float64(), "a": nw.Int16()},
            False,
            False,
            does_not_raise(),
        ),
        (
            {"a": nw.Int32(), "b": nw.Float32()},
            {"b": nw.Float64(), "a": nw.Int16()},
            True,
            False,
            _assertion_error("dtypes do not match"),
        ),
        (
            {"a": nw.Int32(), "b": nw.Float32()},
            {"b": nw.Float64(), "a": nw.Int16()},
            False,
            True,
            _assertion_error("columns are not in the same order"),
        ),
        (
            {"a": nw.Int32(), "b": nw.Float32()},
            {"b": nw.Float64(), "a": nw.Int16()},
            True,
            True,
            _assertion_error("columns are not in the same order"),
        ),
        # Different columns (left not in right)
        (
            {"a": nw.Int32(), "z": nw.String()},
            {"b": nw.Float64()},
            True,
            True,
            _assertion_error("['a', 'z'] in left, but not in right"),
        ),
        # Different columns (right not in left)
        (
            {"z": nw.String()},
            {"z": nw.String(), "b": nw.Float64()},
            True,
            True,
            _assertion_error("['b'] in right, but not in left"),
        ),
    ],
)
def test_check_schema_mismatch(
    constructor: Constructor,
    left_schema: IntoSchema,
    right_schema: IntoSchema,
    *,
    check_dtypes: bool,
    check_column_order: bool,
    context: AbstractContextManager[Any],
) -> None:
    data = {"a": [1, 2, 3], "b": [4.5, 6.7, 8.9], "z": ["foo", "bar", "baz"]}
    left = nw.from_native(constructor(data)).select(
        nw.col(name).cast(dtype) for name, dtype in left_schema.items()
    )
    right = nw.from_native(constructor(data)).select(
        nw.col(name).cast(dtype) for name, dtype in right_schema.items()
    )

    with context:
        assert_frame_equal(
            left, right, check_column_order=check_column_order, check_dtypes=check_dtypes
        )


def test_height_mismatch(constructor: Constructor) -> None:
    left = nw.from_native(constructor({"a": [1, 2, 3]}))
    right = nw.from_native(constructor({"a": [1, 3]}))

    with _assertion_error("height (row count) mismatch"):
        assert_frame_equal(left, right)


@pytest.mark.parametrize("check_row_order", [True, False])
def test_check_row_order(
    constructor: Constructor, request: pytest.FixtureRequest, *, check_row_order: bool
) -> None:
    if "pandas" in str(constructor):  # pragma: no cover
        if PANDAS_VERSION < (2, 2):
            reason = "Pandas too old for nested dtypes"
            pytest.skip(reason=reason)
        pytest.importorskip("pyarrow")

    if "dask" in str(constructor):
        reason = "Unsupported List type"
        request.applymarker(pytest.mark.xfail(reason=reason))

    data = {"a": [1, 2], "b": [["x", "y"], ["x", "z"]]}

    b_expr = nw.col("b").cast(nw.List(nw.String()))
    left = (
        nw.from_native(constructor(data)).with_columns(b_expr).sort("a", descending=False)
    )
    right = (
        nw.from_native(constructor(data)).with_columns(b_expr).sort("a", descending=True)
    )

    context = (
        _assertion_error('value mismatch for column "a"')
        if check_row_order and left.implementation in GUARANTEES_ROW_ORDER
        else does_not_raise()
    )

    with context:
        assert_frame_equal(left, right, check_row_order=check_row_order)


def test_check_row_order_nested_only(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if "pandas" in str(constructor):  # pragma: no cover
        if PANDAS_VERSION < (2, 2):
            reason = "Pandas too old for nested dtypes"
            pytest.skip(reason=reason)
        pytest.importorskip("pyarrow")

    if "dask" in str(constructor):
        reason = "Unsupported List type"
        request.applymarker(pytest.mark.xfail(reason=reason))

    data = {"b": [["x", "y"], ["x", "z"]]}

    b_expr = nw.col("b").cast(nw.List(nw.String()))
    left = nw.from_native(constructor(data)).select(b_expr)

    msg = "`check_row_order=False` is not supported (yet) with only nested data type."
    with pytest.raises(NotImplementedError, match=re.escape(msg)):
        assert_frame_equal(left, left, check_row_order=False)


def test_self_equal(constructor: Constructor, testing_data: Data) -> None:
    """Test that a dataframe is equal to itself, including nested dtypes with nulls.

    We are dropping columns which type is unsupported by _some_ backend.
    """
    cols_to_drop = ("categorical", "enum", "duration", "struct", "time")

    if "pandas" in str(constructor) and PANDAS_VERSION < (2, 2):  # pragma: no cover
        reason = "Pandas too old for nested dtypes"
        pytest.skip(reason=reason)

    _data = {k: v for k, v in testing_data.items() if k not in cols_to_drop}
    df = nw.from_native(constructor(_data))
    assert_frame_equal(df, df)

    # Keep to cover early return when no rows present but same schema
    assert_frame_equal(df.head(0), df.head(0))



from __future__ import annotations

from typing import Any

import pytest

import narwhals as nw
from narwhals.exceptions import ColumnNotFoundError, InvalidOperationError
from tests.utils import Constructor, ConstructorEager, assert_equal_data

data = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8.0, 9.0]}


@pytest.mark.parametrize(
    ("predicates", "expected"),
    [
        ((nw.col("a") > 1,), {"a": [3, 2], "b": [4, 6], "z": [8.0, 9.0]}),
        ((nw.col("a") > 1, nw.col("z") < 9.0), {"a": [3], "b": [4], "z": [8.0]}),
    ],
)
def test_filter_with_expr_predicates(
    constructor: Constructor,
    predicates: tuple[nw.Expr, ...],
    expected: dict[str, list[Any]],
) -> None:
    df = nw.from_native(constructor(data))
    result = df.filter(*predicates)
    assert_equal_data(result, expected)


def test_filter_with_series_predicates(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = df.filter(df["a"] > 1)
    expected = {"a": [3, 2], "b": [4, 6], "z": [8.0, 9.0]}
    assert_equal_data(result, expected)

    result = df.filter(df["a"] > 1, df["b"] < 6)
    expected = {"a": [3], "b": [4], "z": [8.0]}
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("predicates", "expected"),
    [
        (([False, True, True],), {"a": [3, 2], "b": [4, 6], "z": [8.0, 9.0]}),
        (([True, True, False], [False, True, True]), {"a": [3], "b": [4], "z": [8.0]}),
    ],
)
def test_filter_with_boolean_list_predicates_eager(
    constructor_eager: ConstructorEager,
    predicates: tuple[list[bool], ...],
    expected: dict[str, list[Any]],
) -> None:
    df = nw.from_native(constructor_eager(data))
    result = df.filter(*predicates)
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    "predicates",
    [
        ([False, True, True],),
        ([True, True, False], [False, True, True]),
        (nw.col("a") > 1, [False, True, True]),
        ([True, True, False], nw.col("z") < 9.0),
    ],
)
def test_filter_with_boolean_list_predicates_lazy(
    constructor: Constructor, predicates: tuple[list[bool] | nw.Expr, ...]
) -> None:
    df = nw.from_native(constructor(data)).lazy()
    with pytest.raises(TypeError, match="not supported with Python boolean masks"):
        df.filter(*predicates)  # type: ignore[arg-type]


def test_filter_raise_on_agg_predicate(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    with pytest.raises(InvalidOperationError):
        df.filter(nw.col("a").max() > 2).lazy().collect()


def test_filter_raise_on_shape_mismatch(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    with pytest.raises((InvalidOperationError, NotImplementedError)):
        df.filter(nw.col("b").unique() > 2).lazy().collect()


def test_filter_with_constrains_only(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result_scalar = df.filter(a=3)
    expected_scalar = {"a": [3], "b": [4], "z": [8.0]}

    assert_equal_data(result_scalar, expected_scalar)

    result_expr = df.filter(a=nw.col("b") // 3)
    expected_expr = {"a": [1, 2], "b": [4, 6], "z": [7.0, 9.0]}

    assert_equal_data(result_expr, expected_expr)


def test_filter_missing_column(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    constructor_id = str(request.node.callspec.id)
    if any(id_ == constructor_id for id_ in ("sqlframe", "pyspark[connect]", "ibis")):
        request.applymarker(pytest.mark.xfail)

    df = nw.from_native(constructor(data))
    if "polars" in str(constructor):
        msg = r"unable to find column \"c\"; valid columns: \[\"a\", \"b\"\, \"z\"\]"
    elif any(id_ == constructor_id for id_ in ("duckdb", "pyspark")):
        msg = r"\n\nHint: Did you mean one of these columns: \['a', 'b', 'z'\]?"
    else:
        msg = (
            r"The following columns were not found: \[.*\]"
            r"\n\nHint: Did you mean one of these columns: \['a', 'b', 'z'\]?"
        )

    if "polars_lazy" in str(constructor) and isinstance(df, nw.LazyFrame):
        with pytest.raises(ColumnNotFoundError, match=msg):
            df.filter(c=5).collect()
    else:
        with pytest.raises(ColumnNotFoundError, match=msg):
            df.filter(c=5)


def test_filter_with_predicates_and_constraints(
    constructor_eager: ConstructorEager,
) -> None:
    # Adapted from https://github.com/narwhals-dev/narwhals/pull/3173/commits/8433b2d75438df98004a3c850ad23628e2376836
    df = nw.from_native(constructor_eager({"a": range(5), "b": [2, 2, 4, 2, 4]}))
    mask = [True, False, True, True, False]
    mask_2 = [True, True, False, True, False]
    expected_mask_only = {"a": [0, 2, 3], "b": [2, 4, 2]}
    expected_mixed = {"a": [0, 3], "b": [2, 2]}

    result = df.filter(mask)
    assert_equal_data(result, expected_mask_only)

    msg = (
        r"unable to find column \"c\"; valid columns: \[\"a\", \"b\"\]"
        if "polars" in str(constructor_eager)
        else (
            r"The following columns were not found: \[.*\]"
            r"\n\nHint: Did you mean one of these columns: \['a', 'b'\]?"
        )
    )
    with pytest.raises(ColumnNotFoundError, match=msg):
        df.filter(mask, c=1, d=2, e=3, f=4, g=5)

    # NOTE: Everything from here is currently undefined
    result = df.filter(mask, b=2)
    assert_equal_data(result, expected_mixed)

    result = df.filter(mask, nw.col("b") == 2)
    assert_equal_data(result, expected_mixed)

    result = df.filter(mask, mask_2)
    assert_equal_data(result, expected_mixed)

    result = df.filter(
        mask, nw.Series.from_iterable("mask", mask_2, backend=df.implementation)
    )
    assert_equal_data(result, expected_mixed)

    result = df.filter(mask, nw.col("b") != 4, b=2)
    assert_equal_data(result, expected_mixed)


def test_filter_multiple_predicates(constructor: Constructor) -> None:
    """https://github.com/pola-rs/polars/blob/a4522d719de940be3ef99d494ccd1cd6067475c6/py-polars/tests/unit/lazyframe/test_lazyframe.py#L175-L202."""
    df = nw.from_native(
        constructor({"a": [1, 1, 1, 2, 2], "b": [1, 1, 2, 2, 2], "c": [1, 1, 2, 3, 4]})
    )

    # multiple predicates
    expected = {"a": [1, 1, 1], "b": [1, 1, 2], "c": [1, 1, 2]}
    for out in (
        df.filter(nw.col("a") == 1, nw.col("b") <= 2),  # positional/splat
        df.filter([nw.col("a") == 1, nw.col("b") <= 2]),  # as list
    ):
        assert_equal_data(out, expected)

    # multiple kwargs
    assert_equal_data(df.filter(a=1, b=2), {"a": [1], "b": [2], "c": [2]})

    # both positional and keyword args
    assert_equal_data(
        df.filter(nw.col("c") < 4, a=2, b=2), {"a": [2], "b": [2], "c": [3]}
    )


def test_filter_string_predicate(constructor: Constructor) -> None:
    """https://github.com/pola-rs/polars/blob/a4522d719de940be3ef99d494ccd1cd6067475c6/py-polars/tests/unit/lazyframe/test_lazyframe.py#L204-L210."""
    df = nw.from_native(
        constructor({"description": ["eq", "gt", "ge"], "predicate": ["==", ">", ">="]})
    )
    expected = {"description": ["eq"], "predicate": ["=="]}
    result = df.filter(predicate="==")
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    "predicates", [(nw.col("z") < 10,), (nw.col("a") > 0, nw.col("b") > 0)]
)
def test_filter_seq_iterable_all_true(constructor: Constructor, predicates: Any) -> None:
    """https://github.com/pola-rs/polars/blob/a4522d719de940be3ef99d494ccd1cd6067475c6/py-polars/tests/unit/lazyframe/test_lazyframe.py#L213-L233."""
    df = nw.from_native(constructor(data))
    predicate = (p for p in predicates)
    assert_equal_data(df.filter(predicate), data)


@pytest.mark.parametrize(
    "predicates", [(nw.col("z") > 10,), (nw.col("a") < 0, nw.col("b") < 0)]
)
def test_filter_seq_iterable_all_false(constructor: Constructor, predicates: Any) -> None:
    df = nw.from_native(constructor(data))
    expected: dict[str, list[Any]] = {"a": [], "b": [], "z": []}
    predicate = (p for p in predicates)
    assert_equal_data(df.filter(predicate), expected)



from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING, cast

import pytest

import narwhals as nw
from tests.utils import Constructor, ConstructorEager, assert_equal_data

if TYPE_CHECKING:
    import dask.dataframe as dd

data = {
    "a": [datetime(2021, 3, 1, 12, 34, 56, 49000), datetime(2020, 1, 2, 2, 4, 14, 715000)]
}


@pytest.mark.parametrize(
    ("attribute", "expected"),
    [
        ("date", [date(2021, 3, 1), date(2020, 1, 2)]),
        ("year", [2021, 2020]),
        ("month", [3, 1]),
        ("day", [1, 2]),
        ("hour", [12, 2]),
        ("minute", [34, 4]),
        ("second", [56, 14]),
        ("millisecond", [49, 715]),
        ("microsecond", [49000, 715000]),
        ("nanosecond", [49000000, 715000000]),
        ("ordinal_day", [60, 2]),
        ("weekday", [1, 4]),
    ],
)
def test_datetime_attributes(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    attribute: str,
    expected: list[int],
) -> None:
    if (
        attribute == "date"
        and "pandas" in str(constructor)
        and "pyarrow" not in str(constructor)
    ):
        request.applymarker(pytest.mark.xfail)
    if attribute == "date" and "cudf" in str(constructor):
        request.applymarker(pytest.mark.xfail)
    if attribute == "nanosecond" and "ibis" in str(constructor):
        request.applymarker(pytest.mark.xfail)

    df = nw.from_native(constructor(data))
    result = df.select(getattr(nw.col("a").dt, attribute)())
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(
    ("attribute", "expected"),
    [
        ("date", [date(2021, 3, 1), date(2020, 1, 2)]),
        ("year", [2021, 2020]),
        ("month", [3, 1]),
        ("day", [1, 2]),
        ("hour", [12, 2]),
        ("minute", [34, 4]),
        ("second", [56, 14]),
        ("millisecond", [49, 715]),
        ("microsecond", [49000, 715000]),
        ("nanosecond", [49000000, 715000000]),
        ("ordinal_day", [60, 2]),
        ("weekday", [1, 4]),
    ],
)
def test_datetime_attributes_series(
    request: pytest.FixtureRequest,
    constructor_eager: ConstructorEager,
    attribute: str,
    expected: list[int],
) -> None:
    if (
        attribute == "date"
        and "pandas" in str(constructor_eager)
        and "pyarrow" not in str(constructor_eager)
    ):
        request.applymarker(pytest.mark.xfail)
    if attribute == "date" and "cudf" in str(constructor_eager):
        request.applymarker(pytest.mark.xfail)

    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = df.select(getattr(df["a"].dt, attribute)())
    assert_equal_data(result, {"a": expected})


def test_datetime_chained_attributes(
    request: pytest.FixtureRequest, constructor_eager: ConstructorEager
) -> None:
    if "pandas" in str(constructor_eager) and "pyarrow" not in str(constructor_eager):
        request.applymarker(pytest.mark.xfail)
    if "modin" in str(constructor_eager) and "pyarrow" not in str(constructor_eager):
        request.applymarker(pytest.mark.xfail)
    if "cudf" in str(constructor_eager):
        request.applymarker(pytest.mark.xfail)

    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = df.select(df["a"].dt.date().dt.year())
    assert_equal_data(result, {"a": [2021, 2020]})

    result = df.select(nw.col("a").dt.date().dt.year())
    assert_equal_data(result, {"a": [2021, 2020]})


def test_to_date(request: pytest.FixtureRequest, constructor: Constructor) -> None:
    if any(
        x in str(constructor)
        for x in (
            "pandas_constructor",
            "pandas_nullable_constructor",
            "cudf",
            "modin_constructor",
        )
    ):
        request.applymarker(pytest.mark.xfail)
    dates = {"a": [datetime(2001, 1, 1), None, datetime(2001, 1, 3)]}
    if "dask" in str(constructor):
        df_dask = cast("dd.DataFrame", constructor(dates))
        df_dask = cast("dd.DataFrame", df_dask.astype({"a": "timestamp[ns][pyarrow]"}))
        df = nw.from_native(df_dask)
    else:
        df = nw.from_native(constructor(dates))
    result = df.select(nw.col("a").dt.date())
    assert result.collect_schema() == {"a": nw.Date}



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(
    part_ds: FrameT,
    partsupp_ds: FrameT,
    nation_ds: FrameT,
    lineitem_ds: FrameT,
    supplier_ds: FrameT,
) -> FrameT:
    var1 = datetime(1994, 1, 1)
    var2 = datetime(1995, 1, 1)
    var3 = "CANADA"
    var4 = "forest"

    query1 = (
        lineitem_ds.filter(nw.col("l_shipdate").is_between(var1, var2, closed="left"))
        .group_by("l_partkey", "l_suppkey")
        .agg((nw.col("l_quantity").sum()).alias("sum_quantity"))
        .with_columns(sum_quantity=nw.col("sum_quantity") * 0.5)
    )
    query2 = nation_ds.filter(nw.col("n_name") == var3)
    query3 = supplier_ds.join(query2, left_on="s_nationkey", right_on="n_nationkey")  # pyright: ignore[reportArgumentType]

    return (
        part_ds.filter(nw.col("p_name").str.starts_with(var4))
        .select("p_partkey")
        .unique("p_partkey")
        .join(partsupp_ds, left_on="p_partkey", right_on="ps_partkey")
        .join(
            query1,  # pyright: ignore[reportArgumentType]
            left_on=["ps_suppkey", "p_partkey"],
            right_on=["l_suppkey", "l_partkey"],
        )
        .filter(nw.col("ps_availqty") > nw.col("sum_quantity"))
        .select("ps_suppkey")
        .unique("ps_suppkey")
        .join(query3, left_on="ps_suppkey", right_on="s_suppkey")  # pyright: ignore[reportArgumentType]
        .select("s_name", "s_address")
        .sort("s_name")
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from tests.utils import ConstructorEager


def test_sample_n(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(
        constructor_eager({"a": [1, 2, 3, 4], "b": ["x", "y", "x", "y"]}), eager_only=True
    )

    result_expr = df.sample(n=2).shape
    expected_expr = (2, 2)
    assert result_expr == expected_expr


def test_sample_fraction(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(
        constructor_eager({"a": [1, 2, 3, 4], "b": ["x", "y", "x", "y"]}), eager_only=True
    )

    result_expr = df.sample(fraction=0.5).shape
    expected_expr = (2, 2)
    assert result_expr == expected_expr


def test_sample_with_seed(constructor_eager: ConstructorEager) -> None:
    size, n = 100, 10
    df = nw.from_native(constructor_eager({"a": range(size)}), eager_only=True)

    r1 = nw.to_native(df.sample(n=n, seed=123))
    r2 = nw.to_native(df.sample(n=n, seed=123))
    r3 = nw.to_native(df.sample(n=n, seed=42))

    assert r1.equals(r2)  # type: ignore[attr-defined]
    assert not r1.equals(r3)  # type: ignore[attr-defined]



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(
    nation_ds: FrameT,
    customer_ds: FrameT,
    line_item_ds: FrameT,
    orders_ds: FrameT,
    supplier_ds: FrameT,
) -> FrameT:
    n1 = nation_ds.filter(nw.col("n_name") == "FRANCE")
    n2 = nation_ds.filter(nw.col("n_name") == "GERMANY")

    var_1 = datetime(1995, 1, 1)
    var_2 = datetime(1996, 12, 31)

    df1 = (
        customer_ds.join(n1, left_on="c_nationkey", right_on="n_nationkey")  # pyright: ignore[reportArgumentType]
        .join(orders_ds, left_on="c_custkey", right_on="o_custkey")
        .rename({"n_name": "cust_nation"})
        .join(line_item_ds, left_on="o_orderkey", right_on="l_orderkey")
        .join(supplier_ds, left_on="l_suppkey", right_on="s_suppkey")
        .join(n2, left_on="s_nationkey", right_on="n_nationkey")  # pyright: ignore[reportArgumentType]
        .rename({"n_name": "supp_nation"})
    )

    df2 = (
        customer_ds.join(n2, left_on="c_nationkey", right_on="n_nationkey")  # pyright: ignore[reportArgumentType]
        .join(orders_ds, left_on="c_custkey", right_on="o_custkey")
        .rename({"n_name": "cust_nation"})
        .join(line_item_ds, left_on="o_orderkey", right_on="l_orderkey")
        .join(supplier_ds, left_on="l_suppkey", right_on="s_suppkey")
        .join(n1, left_on="s_nationkey", right_on="n_nationkey")  # pyright: ignore[reportArgumentType]
        .rename({"n_name": "supp_nation"})
    )

    return (
        nw.concat([df1, df2])
        .filter(nw.col("l_shipdate").is_between(var_1, var_2))
        .with_columns(
            (nw.col("l_extendedprice") * (1 - nw.col("l_discount"))).alias("volume")
        )
        .with_columns(nw.col("l_shipdate").dt.year().alias("l_year"))
        .group_by("supp_nation", "cust_nation", "l_year")
        .agg(nw.sum("volume").alias("revenue"))
        .sort(by=["supp_nation", "cust_nation", "l_year"])
    )



from __future__ import annotations

import pytest

pytest.importorskip("numpy")
import numpy as np

import narwhals as nw
from tests.utils import (
    PANDAS_VERSION,
    PYARROW_VERSION,
    ConstructorEager,
    assert_equal_data,
)


def test_array_dunder(
    request: pytest.FixtureRequest, constructor_eager: ConstructorEager
) -> None:
    if "pyarrow_table" in str(constructor_eager) and PYARROW_VERSION < (
        16,
        0,
        0,
    ):  # pragma: no cover
        request.applymarker(pytest.mark.xfail)

    s = nw.from_native(constructor_eager({"a": [1, 2, 3]}), eager_only=True)["a"]
    result = s.__array__()
    np.testing.assert_array_equal(result, np.array([1, 2, 3], dtype="int64"))


def test_array_dunder_with_dtype(
    request: pytest.FixtureRequest, constructor_eager: ConstructorEager
) -> None:
    if "pyarrow_table" in str(constructor_eager) and PYARROW_VERSION < (
        16,
        0,
        0,
    ):  # pragma: no cover
        request.applymarker(pytest.mark.xfail)

    s = nw.from_native(constructor_eager({"a": [1, 2, 3]}), eager_only=True)["a"]
    result = s.__array__(object)
    np.testing.assert_array_equal(result, np.array([1, 2, 3], dtype=object))


def test_array_dunder_with_copy(
    request: pytest.FixtureRequest, constructor_eager: ConstructorEager
) -> None:
    if "pyarrow_table" in str(constructor_eager) and PYARROW_VERSION < (
        16,
        0,
        0,
    ):  # pragma: no cover
        request.applymarker(pytest.mark.xfail)

    s = nw.from_native(constructor_eager({"a": [1, 2, 3]}), eager_only=True)["a"]
    result = s.__array__(copy=True)
    np.testing.assert_array_equal(result, np.array([1, 2, 3], dtype="int64"))
    if "pandas_constructor" in str(constructor_eager) and PANDAS_VERSION < (3,):
        # If it's pandas, we know that `copy=False` definitely took effect.
        # So, let's check it!
        result = s.__array__(copy=False)
        np.testing.assert_array_equal(result, np.array([1, 2, 3], dtype="int64"))
        result[0] = 999
        assert_equal_data({"a": s}, {"a": [999, 2, 3]})



from __future__ import annotations

import narwhals as nw
from tests.utils import Constructor, ConstructorEager, assert_equal_data


def test_str_pad_start_series(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(
        constructor_eager({"a": ["foo", "longer_foo", "longest_fooooooo", "hi", None]})
    )

    result = {
        "padded": df["a"].str.pad_start(10),
        "padded_len": df["a"].str.pad_start(10).str.len_chars(),
    }
    expected = {
        "padded": ["       foo", "longer_foo", "longest_fooooooo", "        hi", None],
        "padded_len": [10, 10, 16, 10, None],
    }

    assert_equal_data(result, expected)


def test_str_pad_start_expr(constructor: Constructor) -> None:
    df = nw.from_native(
        constructor({"a": ["foo", "longer_foo", "longest_fooooooo", "hi", None]})
    )

    result = df.select(
        nw.col("a").str.pad_start(10).alias("padded"),
        nw.col("a").str.pad_start(10).str.len_chars().alias("padded_len"),
    )
    expected = {
        "padded": ["       foo", "longer_foo", "longest_fooooooo", "        hi", None],
        "padded_len": [10, 10, 16, 10, None],
    }

    assert_equal_data(result, expected)


def test_str_pad_end_series(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(
        constructor_eager({"a": ["foo", "longer_foo", "longest_fooooooo", "hi", None]})
    )

    result = {
        "padded": df["a"].str.pad_end(10),
        "padded_len": df["a"].str.pad_end(10).str.len_chars(),
    }
    expected = {
        "padded": ["foo       ", "longer_foo", "longest_fooooooo", "hi        ", None],
        "padded_len": [10, 10, 16, 10, None],
    }

    assert_equal_data(result, expected)


def test_str_pad_end_expr(constructor: Constructor) -> None:
    df = nw.from_native(
        constructor({"a": ["foo", "longer_foo", "longest_fooooooo", "hi", None]})
    )

    result = df.select(
        nw.col("a").str.pad_end(10).alias("padded"),
        nw.col("a").str.pad_end(10).str.len_chars().alias("padded_len"),
    )
    expected = {
        "padded": ["foo       ", "longer_foo", "longest_fooooooo", "hi        ", None],
        "padded_len": [10, 10, 16, 10, None],
    }

    assert_equal_data(result, expected)


def test_pad_start_unicode_expr(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": ["Café", "345", "東京", None]}))

    result = df.select(nw.col("a").str.pad_start(6, "日"))
    expected = {"a": ["日日Café", "日日日345", "日日日日東京", None]}

    assert_equal_data(result, expected)


def test_pad_start_unicode_series(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager({"a": ["Café", "345", "東京", None]}))

    result = {"a": df["a"].str.pad_start(6, "日")}
    expected = {"a": ["日日Café", "日日日345", "日日日日東京", None]}

    assert_equal_data(result, expected)


def test_pad_end_unicode_expr(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": ["Café", "345", "東京", None]}))

    result = df.select(nw.col("a").str.pad_end(6, "日"))
    expected = {"a": ["Café日日", "345日日日", "東京日日日日", None]}

    assert_equal_data(result, expected)


def test_pad_end_unicode_series(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager({"a": ["Café", "345", "東京", None]}))

    result = {"a": df["a"].str.pad_end(6, "日")}
    expected = {"a": ["Café日日", "345日日日", "東京日日日日", None]}

    assert_equal_data(result, expected)



from __future__ import annotations

from contextlib import nullcontext as does_not_raise

import pytest

import narwhals as nw
from tests.utils import Constructor, ConstructorEager, assert_equal_data

data = {"a": [1, 3, 3], "b": [1, 2, 3], "c": [1, None, 1]}


@pytest.mark.parametrize(
    ("output_name", "a", "b", "expected_corr"),
    [
        ("a", "a", "b", 0.87),
        ("a", "a", "c", None),
        ("b", nw.col("a").alias("b"), nw.col("b").alias("c"), 0.87),
        ("a", nw.col("a") * nw.col("b"), nw.col("b") - 0.5, 0.99),
    ],
)
def test_corr_expr(
    constructor: Constructor,
    request: pytest.FixtureRequest,
    output_name: str,
    a: str | nw.Expr,
    b: str | nw.Expr,
    expected_corr: float,
) -> None:
    if "pyspark" in str(constructor) and expected_corr is None:
        request.applymarker(
            pytest.skip(reason="Pyspark corr function does not allow None values")
        )
    df = nw.from_native(constructor(data))
    result = df.select(nw.corr(a, b).round(2))
    expected = {output_name: [expected_corr]}
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("output_name", "a", "b", "expected_corr"),
    [
        ("a", "a", "b", 0.87),
        ("a", "a", "c", None),
        ("b", nw.col("a").alias("b"), nw.col("b").alias("c"), 0.87),
        ("a", nw.col("a") * nw.col("b"), nw.col("b") - 0.5, 1),
    ],
)
def test_corr_expr_spearman(
    constructor: Constructor,
    output_name: str,
    a: str | nw.Expr,
    b: str | nw.Expr,
    expected_corr: float,
) -> None:
    context = (
        does_not_raise()
        if any(x in str(constructor) for x in ("pandas", "polars", "modin", "cudf"))
        else pytest.raises(NotImplementedError)
    )
    df = nw.from_native(constructor(data))
    with context:
        result = df.select(nw.corr(a, b, method="spearman").round(2))
        expected = {output_name: [expected_corr]}
        assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("output_name", "a", "b", "expected_corr"),
    [("a", "a", "b", 0.87), ("a", "a", "c", None)],
)
def test_corr_series(
    constructor_eager: ConstructorEager,
    request: pytest.FixtureRequest,
    output_name: str,
    a: str,
    b: str,
    expected_corr: float,
) -> None:
    if "pyspark" in str(constructor_eager) and expected_corr is None:
        request.applymarker(
            pytest.skip(reason="Pyspark corr function does not allow None values")
        )
    df = nw.from_native(constructor_eager(data))
    result = df.select(nw.corr(df[a], df[b]).round(2))
    expected = {output_name: [expected_corr]}
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("output_name", "a", "b", "expected_corr"),
    [("a", "a", "b", 0.87), ("a", "a", "c", None)],
)
def test_corr_series_spearman(
    constructor_eager: ConstructorEager,
    request: pytest.FixtureRequest,
    output_name: str,
    a: str,
    b: str,
    expected_corr: float,
) -> None:
    if "pyspark" in str(constructor_eager) and expected_corr is None:
        request.applymarker(
            pytest.skip(reason="Pyspark corr function does not allow None values")
        )
    context = (
        does_not_raise()
        if any(
            x in str(constructor_eager)
            for x in ("pandas", "polars", "modin", "cudf", "pyarrow")
        )
        else pytest.raises(NotImplementedError)
    )
    df = nw.from_native(constructor_eager(data))
    with context:
        result = df.select(nw.corr(df[a], df[b]).round(2))
        expected = {output_name: [expected_corr]}
        assert_equal_data(result, expected)
